# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import copy
import itertools

from pyomo.common import DeveloperError
from pyomo.common.collections import Sequence

from pyomo.core.base.enums import SortComponents
from pyomo.core.base.global_set import UnindexedComponent_index


class IndexedComponent_slice:
    """Special class for slicing through hierarchical component trees

    The basic concept is to interrupt the normal slice generation
    procedure to return a specialized iterable class (this object).  This
    object supports simple getitem / getattr / call methods and caches
    them until it is time to actually iterate through the slice.  We
    then walk down the cached names / indices and resolve the final
    objects during the iteration process.  This works because all the
    calls to __getitem__ / __getattr__ / __call__ happen *before* the
    call to __iter__()
    """

    ATTR_MASK = 8
    ITEM_MASK = 16
    CALL_MASK = 32
    GET_MASK = 1
    SET_MASK = 2
    DEL_MASK = 4

    slice_info = 0
    get_attribute = ATTR_MASK | GET_MASK
    set_attribute = ATTR_MASK | SET_MASK
    del_attribute = ATTR_MASK | DEL_MASK
    get_item = ITEM_MASK | GET_MASK
    set_item = ITEM_MASK | SET_MASK
    del_item = ITEM_MASK | DEL_MASK
    call = CALL_MASK

    def __init__(self, component, fixed=None, sliced=None, ellipsis=None):
        """A "slice" over an _IndexedComponent hierarchy

        This class has two forms for the constructor.  The first form is
        the standard constructor that takes a base component and
        indexing information.  This form takes

           IndexedComponent_slice(component, fixed, sliced, ellipsis)

        The second form is a "copy constructor" that is used internally
        when building up the "call stack" for the hierarchical slice.  The
        copy constructor takes an IndexedComponent_slice and an
        optional "next term" in the slice construction (from get/set/del
        item/attr or call):

           IndexedComponent_slice(slice, next_term=None)

        Parameters
        ----------
        component: IndexedComponent
            The base component for this slice

        fixed: dict
            A dictionary indicating the fixed indices of component,
            mapping index position to value

        sliced: dict
            A dictionary indicating the sliced indices of component
            mapping the index position to the (python) slice object

        ellipsis: int
            The position of the ellipsis in the initial component slice

        """
        # Note that because we use a custom __setattr__, we need to
        # define actual instance attributes using the base class
        # __setattr__.
        set_attr = super(IndexedComponent_slice, self).__setattr__
        if type(component) is IndexedComponent_slice:
            # Copy constructor
            _len = component._len
            # For efficiency, we will only duplicate the call stack
            # list if this instance is not point to the end of the list.
            if _len == len(component._call_stack):
                set_attr('_call_stack', component._call_stack)
            else:
                set_attr('_call_stack', component._call_stack[:_len])
            set_attr('_len', _len)
            if fixed is not None:
                self._call_stack.append(fixed)
                self._len += 1
            set_attr(
                'call_errors_generate_exceptions',
                component.call_errors_generate_exceptions,
            )
            set_attr(
                'key_errors_generate_exceptions',
                component.key_errors_generate_exceptions,
            )
            set_attr(
                'attribute_errors_generate_exceptions',
                component.attribute_errors_generate_exceptions,
            )
        else:
            # Normal constructor
            set_attr(
                '_call_stack',
                [
                    (
                        IndexedComponent_slice.slice_info,
                        (component, fixed, sliced, ellipsis),
                    )
                ],
            )
            set_attr('_len', 1)
            # Since this is an object, users may change these flags
            # between where they declare the slice and iterate over it.
            set_attr('call_errors_generate_exceptions', True)
            set_attr('key_errors_generate_exceptions', True)
            set_attr('attribute_errors_generate_exceptions', True)

    def __getstate__(self):
        """Serialize this object.

        In general, we would not need to implement this (the object does
        not leverage ``__slots__``).  However, because we have a
        "blanket" implementation of :py:meth:`__getattr__`, we need to
        explicitly implement these to avoid "accidentally" extending or
        evaluating this slice."""
        return {k: getattr(self, k) for k in self.__dict__}

    def __setstate__(self, state):
        """Deserialize the state into this object."""
        set_attr = super(IndexedComponent_slice, self).__setattr__
        for k, v in state.items():
            set_attr(k, v)

    def __deepcopy__(self, memo):
        """Deepcopy this object (leveraging :py:meth:`__getstate__`)"""
        ans = memo[id(self)] = self.__class__.__new__(self.__class__)
        ans.__setstate__(copy.deepcopy(self.__getstate__(), memo))
        return ans

    def __iter__(self):
        """Return an iterator over this slice"""
        return _IndexedComponent_slice_iter(self)

    def __getattr__(self, name):
        """Override the "." operator to defer resolution until iteration.

        Creating a slice of a component returns a
        IndexedComponent_slice object.  Subsequent attempts to resolve
        attributes hit this method.
        """
        return IndexedComponent_slice(
            self, (IndexedComponent_slice.get_attribute, name)
        )

    def __setattr__(self, name, value):
        """Override the "." operator implementing attribute assignment

        This supports notation similar to:

            del model.b[:].c.x = 5

        and immediately evaluates the slice.
        """
        # Don't overload any pre-existing attributes
        if name in self.__dict__:
            return super(IndexedComponent_slice, self).__setattr__(name, value)

        # Immediately evaluate the slice and set the attributes
        for i in IndexedComponent_slice(
            self, (IndexedComponent_slice.set_attribute, name, value)
        ):
            pass
        return None

    def __getitem__(self, idx):
        """Override the "[]" operator to defer resolution until iteration.

        Creating a slice of a component returns a
        IndexedComponent_slice object.  Subsequent attempts to query
        items hit this method.
        """
        return IndexedComponent_slice(self, (IndexedComponent_slice.get_item, idx))

    def __setitem__(self, idx, val):
        """Override the "[]" operator for setting item values.

        This supports notation similar to:

            model.b[:].c.x[1,:] = 5

        and immediately evaluates the slice.
        """
        # Immediately evaluate the slice and set the attributes
        for i in IndexedComponent_slice(
            self, (IndexedComponent_slice.set_item, idx, val)
        ):
            pass
        return None

    def __delitem__(self, idx):
        """Override the "del []" operator for deleting item values.

        This supports notation similar to:

            del model.b[:].c.x[1,:]

        and immediately evaluates the slice.
        """
        # Immediately evaluate the slice and set the attributes
        for i in IndexedComponent_slice(self, (IndexedComponent_slice.del_item, idx)):
            pass
        return None

    def __call__(self, *args, **kwds):
        """Special handling of the "()" operator for component slices.

        Creating a slice of a component returns a IndexedComponent_slice
        object.  Subsequent attempts to call items hit this method.  We
        handle the __call__ method separately based on the item (identifier
        immediately before the "()") being called:

        - if the item was 'component', then we defer resolution of this call
        until we are actually iterating over the slice.  This allows users
        to do operations like `m.b[:].component('foo').bar[:]`

        - if the item is anything else, then we will immediately iterate over
        the slice and call the item.  This allows "vector-like" operations
        like: `m.x[:,1].fix(0)`.
        """
        # There is a weird case in pypy3.6-7.2.0 where __name__ gets
        # called after retrieving an attribute that will be called.  I
        # don't know why that happens, but we will trap it here and
        # remove the getattr(__name__) from the call stack.
        _len = self._len
        if (
            self._call_stack[_len - 1][0] == IndexedComponent_slice.get_attribute
            and self._call_stack[_len - 1][1] == '__name__'
        ):
            self._len -= 1

        ans = IndexedComponent_slice(self, (IndexedComponent_slice.call, args, kwds))
        # Because we just duplicated the slice and added a new entry, we
        # know that the _len == len(_call_stack)
        if ans._call_stack[-2][1] == 'component':
            return ans
        else:
            # Note: simply calling "list(self)" results in infinite
            # recursion in python2.6
            return list(i for i in ans)

    @classmethod
    def _getitem_args_to_str(cls, args):
        for i, v in enumerate(args):
            if v is Ellipsis:
                args[i] = '...'
            elif type(v) is slice:
                args[i] = (
                    (repr(v.start) if v.start is not None else '')
                    + ':'
                    + (repr(v.stop) if v.stop is not None else '')
                    + (':%r' % v.step if v.step is not None else '')
                )
            else:
                args[i] = repr(v)
        return '[' + ', '.join(args) + ']'

    def __str__(self):
        ans = ''
        for level in self._call_stack:
            if level[0] == IndexedComponent_slice.slice_info:
                ans += level[1][0].name
                if level[1][1] is not None:
                    tmp = dict(level[1][1])
                    tmp.update(level[1][2])
                    if level[1][3] is not None:
                        tmp[level[1][3]] = Ellipsis
                    ans += self._getitem_args_to_str([tmp[i] for i in sorted(tmp)])
            elif level[0] & IndexedComponent_slice.ITEM_MASK:
                if isinstance(level[1], Sequence):
                    tmp = list(level[1])
                else:
                    tmp = [level[1]]
                ans += self._getitem_args_to_str(tmp)
            elif level[0] & IndexedComponent_slice.ATTR_MASK:
                ans += '.' + level[1]
            elif level[0] & IndexedComponent_slice.CALL_MASK:
                ans += (
                    '('
                    + ', '.join(
                        itertools.chain(
                            (repr(_) for _ in level[1]),
                            ('%s=%r' % kv for kv in level[2].items()),
                        )
                    )
                    + ')'
                )
            if level[0] & IndexedComponent_slice.SET_MASK:
                ans += ' = %r' % (level[2],)
            elif level[0] & IndexedComponent_slice.DEL_MASK:
                ans = 'del ' + ans
        return ans

    def __hash__(self):
        return hash(tuple(_freeze(x) for x in self._call_stack[: self._len]))

    def __eq__(self, other):
        if other is self:
            return True
        if type(other) is not IndexedComponent_slice:
            return False
        return tuple(_freeze(x) for x in self._call_stack[: self._len]) == tuple(
            _freeze(x) for x in other._call_stack[: other._len]
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def duplicate(self):
        ans = IndexedComponent_slice(self)
        ans._call_stack = ans._call_stack[: ans._len]
        return ans

    def index_wildcard_keys(self, sort):
        _iter = _IndexedComponent_slice_iter(self, iter_over_index=True, sort=sort)
        return (_iter.get_last_index_wildcards() for _ in _iter)

    def wildcard_keys(self, sort=SortComponents.UNSORTED):
        _iter = _IndexedComponent_slice_iter(self, sort=sort)
        return (_iter.get_last_index_wildcards() for _ in _iter)

    def wildcard_values(self, sort=SortComponents.UNSORTED):
        """Return an iterator over this slice"""
        return _IndexedComponent_slice_iter(self, sort=sort)

    def wildcard_items(self, sort=SortComponents.UNSORTED):
        _iter = _IndexedComponent_slice_iter(self, sort=sort)
        return ((_iter.get_last_index_wildcards(), _) for _ in _iter)

    def expanded_keys(self):
        _iter = self.__iter__()
        return (_iter.get_last_index() for _ in _iter)

    def expanded_items(self):
        _iter = self.__iter__()
        return ((_iter.get_last_index(), _) for _ in _iter)


def _freeze(info):
    if info[0] == IndexedComponent_slice.slice_info:
        return (
            info[0],
            id(info[1][0]),  # id of the Component
            tuple(info[1][1].items()),  # {idx: value} for fixed
            tuple(info[1][2].keys()),  # {idx: slice} for slices
            info[1][3],  # ellipsis index
        )
    elif info[0] & IndexedComponent_slice.ITEM_MASK:
        if type(info[1]) is not tuple:
            index = (info[1],)
        else:
            index = info[1]
        return (
            info[0],
            tuple((x.start, x.stop, x.step) if type(x) is slice else x for x in index),
            info[2:],
        )
    else:
        return info


class _slice_generator:
    """Utility (iterator) for generating the elements of one slice

    Iterate through the component index and yield the component data
    values that match the slice template.
    """

    def __init__(self, component, fixed, sliced, ellipsis, iter_over_index, sort):
        self.component = component
        self.fixed = fixed
        self.sliced = sliced
        self.ellipsis = ellipsis
        self.iter_over_index = iter_over_index

        # Cache for the most recent index returned. This is used to
        # iterate over keys of the slice (for instance, in a
        # _ReferenceDict).
        self.last_index = ()

        self.tuplize_unflattened_index = (
            len(list(self.component.index_set().subsets())) <= 1
        )

        if fixed is None and sliced is None and ellipsis is None:
            # This is a slice rooted at a concrete component.  This is
            # unusual (i.e., it is generated programmatically and not
            # through the normal slice/ellipsis notation).  We will mock
            # up a "generator" so that the slice can be cleanly iterated
            # over
            self.explicit_index_count = 0
            self.component_iter = _NotIterable
            return

        self.explicit_index_count = len(fixed) + len(sliced)
        if iter_over_index and component.index_set().isfinite():
            # This should be used to iterate over all the potential
            # indices of a sparse IndexedComponent.
            if SortComponents.SORTED_INDICES in sort:
                self.component_iter = component.index_set().sorted_iter()
            elif SortComponents.ORDERED_INDICES in sort:
                self.component_iter = component.index_set().ordered_iter()
            else:
                self.component_iter = iter(component.index_set())
        else:
            # The default behavior is to iterate over the component.
            self.component_iter = component.keys(sort)

    def next(self):
        """__next__() iterator for Py2 compatibility"""
        return self.__next__()

    def __next__(self):
        # We have to defer this import to here to resolve circular
        # imports.  Ideally, we would move normalize_index to another
        # module to resolve this.
        from .indexed_component import normalize_index

        if self.component_iter is _NotIterable:
            # Special case handling for "slices" rooted at concrete
            # components.  We will replace the iterator with an "empty"
            # iterator so that the next call will raise StopIteration
            self.component_iter = iter(())
            return self.component

        while 1:
            # Note: running off the end of the underlying iterator will
            # generate a StopIteration exception that will propagate up
            # and end this iterator.
            index = next(self.component_iter)

            # We want a tuple of indices, so convert scalars to tuples
            if normalize_index.flatten:
                _idx = index if type(index) is tuple else (index,)
            elif self.tuplize_unflattened_index:
                _idx = (index,)
            else:
                _idx = index

            # Verify the number of indices: if there is a wildcard
            # slice, then there must be enough indices to at least match
            # the fixed indices.  Without the wildcard slice (ellipsis),
            # the number of indices must match exactly.
            if self.ellipsis is not None:
                if self.explicit_index_count > len(_idx):
                    continue
            elif len(_idx) != self.explicit_index_count:
                continue

            valid = True
            for key, val in self.fixed.items():
                # If this index of the component does not match all
                # the specified fixed indices, don't return anything.
                if not val == _idx[key]:
                    valid = False
                    break
            if valid:
                # Remember the index tuple corresponding to the last
                # component data returned by this iterator. In this way
                # we can use the cached indices to iterate over "indices"
                # of a slice.
                #
                # last_index is the most recent index encountered, not
                # the last index that will ever be encountered.
                self.last_index = _idx

                # Note: it is important to use __getitem__, as the
                # derived class may implement a non-standard storage
                # mechanism (e.g., Param)
                if (not self.iter_over_index) or index in self.component:
                    # If iter_over_index is False, we are iterating over
                    # the component ("filled-in" indices only).  Since
                    # `advance_iter` was called on the component iter,
                    # we already know index is in self.component.
                    return self.component[index]
                else:
                    # If iter_over_index is True, we need to return
                    # something even when index is not actually in
                    # self.component.  We will (arbitrarily) return
                    # None.
                    return None


# Backwards compatibility
_IndexedComponent_slice = IndexedComponent_slice


# Mock up a callable object with a "check_complete" method
def _advance_iter(_iter):
    return next(_iter)


def _advance_iter_check_complete():
    pass


_advance_iter.check_complete = _advance_iter_check_complete


# A dummy class that we can use as a named entity below
class _NotIterable:
    pass


class _IndexedComponent_slice_iter:
    def __init__(
        self,
        component_slice,
        advance_iter=_advance_iter,
        iter_over_index=False,
        sort=False,
    ):
        # _iter_stack holds a list of elements X where X is either a
        # _slice_generator iterator (if this level in the hierarchy is a
        # slice) or None (if this level is either a SimpleComponent,
        # attribute, method, or is explicitly indexed).
        self._slice = component_slice
        self.advance_iter = advance_iter
        self._iter_over_index = iter_over_index
        self._sort = SortComponents(sort)
        call_stack = self._slice._call_stack
        call_stack_len = self._slice._len
        self._iter_stack = [None] * call_stack_len
        # Initialize the top of the `_iter_stack` (deepest part of the
        # model hierarchy):
        if call_stack[0][0] == IndexedComponent_slice.slice_info:
            # The root of the _iter_stack is a generator for the
            # "highest-level slice" (slice closest to the model() block)
            self._iter_stack[0] = _slice_generator(
                *call_stack[0][1],
                iter_over_index=self._iter_over_index,
                sort=self._sort,
            )
            # call_stack[0][1] is a (fixed, sliced, ellipsis) tuple, where
            # fixed and sliced are dicts.
        elif call_stack[0][0] == IndexedComponent_slice.set_item:
            # This is a special case that happens when calling
            # `_ReferenceDict.__setitem__` when the call stack consists
            # of only a `set_item` entry.  We need to initialize the
            # root of _iter_stack to something other than None (so it
            # doesn't immediately get "popped" off.  However, set_item
            # is not an iterable thing, so we will will use a type flag
            # to signal this case to __next__below.
            assert call_stack_len == 1
            self._iter_stack[0] = _NotIterable  # Something not None
        else:
            raise DeveloperError(
                "Unexpected call_stack flag encountered: %s" % call_stack[0][0]
            )

    def __iter__(self):
        """This class implements the iterator API"""
        return self

    def next(self):
        """__next__() iterator for Py2 compatibility"""
        return self.__next__()

    def __next__(self):
        """Return the next element in the slice."""
        # In each call to this function, idx will initially point
        # to the bottom of the stack.
        #
        # NOTE: We refer to this stack as growing "downward", just like
        # the model hierarchy to which it refers.
        idx = len(self._iter_stack) - 1
        while True:
            # Flush out any non-slice levels.  Since we initialize
            # _iter_stack with None, in the first call this will
            # immediately walk up to the beginning of the _iter_stack
            #
            # On subsequent calls, we will walk up only as far as the
            # "deepest" active (non-exhausted) iterator. Higher-level
            # iterators could still be active as well, but those index
            # values will remain constant until we have exhausted this
            # "deepest" iterator.
            while self._iter_stack[idx] is None:
                idx -= 1
            # Get the next element in the deepest iterator (active slice)
            try:
                if self._iter_stack[idx] is _NotIterable:
                    # This happens when attempting a `set_item` call on
                    # a `_ReferenceDict` whose slice consists of only a
                    # `slice_info` entry.
                    # E.g.
                    #     ref = Reference(m.x[:])
                    #     ref._data[1] = 2
                    # but not
                    #     ref = Reference(m.b[:].x[:])
                    #     ref._data['a',1] = 2
                    #
                    _comp = self._slice._call_stack[0][1][0]
                    # _comp is the component in the slice_info entry
                    # of the call stack
                else:
                    # Advance the "deepest active iterator"
                    _comp = self.advance_iter(self._iter_stack[idx])
                    # Note that if we are looking for a specific
                    # wildcard index, that data is stored in
                    # advance_iter() and will be automatically inserted.
                    #
                    # _comp is "local" part of the component we're
                    # looking for.  The rest of the component will be
                    # located using the remainder of the iter stack.
                    #
                    # Note that _comp is actually a component data, because
                    # the _slice_generator (_iter_stack[idx]) returns
                    # component datas rather than indices.
                    # The _slice_generator is able to know about its
                    # component because it was created from a "higher-
                    # level" component/slice in the call/iter stack.
                    # A higher-level iterator may still be active, and
                    # this _slice_generator will need to be regenerated
                    # when/if that iterator is advanced.
                    idx += 1
            except StopIteration:
                # We have exhausted the iterator at this level of the
                # stack
                if not idx:
                    # Top-level iterator is done.  We are done.
                    # (This is how the infinite loop terminates!)
                    raise
                # Reset the _slice_generator to None so that the next
                # iteration will walk up to - and advance - the
                # "next-highest level" iterator.
                self._iter_stack[idx] = None
                # Trivial optimization: we now know that the
                # _iter_stack[idx] is None, so we can preemptively
                # decrement idx in preparatioon for the next iteration
                # of this loop.
                idx -= 1
                continue

            # Walk to the end of the iter/call stacks, constructing a
            # component to return along the way.  The _iter_stack for
            # all of these levels from idx to the end of the list are
            # known to be None at this point.
            while idx < self._slice._len:
                _call = self._slice._call_stack[idx]
                if _call[0] == IndexedComponent_slice.get_attribute:
                    try:
                        # Attach attribute to our current component:
                        _comp = getattr(_comp, _call[1])
                    except AttributeError:
                        # Since we are slicing, we may only be interested in
                        # things that match.  We will allow users to
                        # (silently) ignore any attribute errors generated
                        # by concrete indices in the slice hierarchy...
                        if (
                            self._slice.attribute_errors_generate_exceptions
                            and not self._iter_over_index
                        ):
                            raise
                        # Break from the inner loop; next action will be to
                        # advance the "highest-level iterator"
                        break
                elif _call[0] == IndexedComponent_slice.get_item:
                    try:
                        # Get the specified index for the current component:
                        _comp = _comp.__getitem__(_call[1])
                    except LookupError:
                        # Since we are slicing, we may only be
                        # interested in things that match.  We will
                        # allow users to (silently) ignore any key
                        # errors generated by concrete indices in the
                        # slice hierarchy...
                        if (
                            self._slice.key_errors_generate_exceptions
                            and not self._iter_over_index
                        ):
                            raise
                        break
                    # If the index defines a slice, add a slice generator
                    # to the iter_stack:
                    if _comp.__class__ is IndexedComponent_slice:
                        # Extract the _slice_generator (for
                        # efficiency... these are always 1-level slices,
                        # so we don't need the overhead of the
                        # IndexedComponent_slice object)
                        assert _comp._len == 1
                        self._iter_stack[idx] = _slice_generator(
                            *_comp._call_stack[0][1],
                            iter_over_index=self._iter_over_index,
                            sort=self._sort,
                        )
                        try:
                            # Advance to get the first component defined
                            # by this slice (so that we have a concrete
                            # context that we can use to descend further
                            # down the model hierarchy):
                            _comp = self.advance_iter(self._iter_stack[idx])
                            # Note that the iterator will remained
                            # cached for subsequent calls to __next__()
                            # (when it will eventually be exhausted).
                        except StopIteration:
                            # We got a slicer, but the slicer doesn't
                            # match anything.  We should break here,
                            # which (due to 'while True' above) will
                            # walk back up to the next iterator and move
                            # on
                            self._iter_stack[idx] = None
                            break
                    else:
                        # `_comp` is a fully qualified component data (i.e.,
                        # not a slice).  Record None in the _iter_stack
                        # so we note that this level in the stack is not
                        # defined by an iterator (i.e., subsequent calls
                        # should immediately "pop" this level off the
                        # stack and proceed to the next higher level.
                        self._iter_stack[idx] = None
                elif _call[0] == IndexedComponent_slice.call:
                    try:
                        # Assume the callable "comp" in our hierarchy
                        # returns a component:
                        _comp = _comp(*(_call[1]), **(_call[2]))
                    except:
                        # Since we are slicing, we may only be
                        # interested in things that match.  We will
                        # allow users to (silently) ignore any key
                        # errors generated by concrete indices in the
                        # slice hierarchy...
                        if (
                            self._slice.call_errors_generate_exceptions
                            and not self._iter_over_index
                        ):
                            raise
                        break
                elif _call[0] == IndexedComponent_slice.set_attribute:
                    # set_attribute should only appear at the deepest
                    # point (end) of the call stack
                    assert idx == self._slice._len - 1
                    try:
                        # set attribute of this component:
                        _comp = setattr(_comp, _call[1], _call[2])
                        # If we want to support __setattr__ with "vector"
                        # arguments, e.g.
                        # `m.b[:].v.value = [1,2,3]` or
                        # `m.b[:].v[1].value = m.b[:].v[0].value`,
                        # this will need to be modified to
                        # cache an iterator over _call[2].
                    except AttributeError:
                        # Since we are slicing, we may only be interested in
                        # things that match.  We will allow users to
                        # (silently) ignore any attribute errors generated
                        # by concrete indices in the slice hierarchy...
                        if self._slice.attribute_errors_generate_exceptions:
                            raise
                        break
                elif _call[0] == IndexedComponent_slice.set_item:
                    # `set_item` must always appear at the deepest
                    # point (end) of the call stack
                    assert idx == self._slice._len - 1
                    # We have a somewhat unusual situation when someone
                    # makes a _ReferenceDict to m.x[:] and then wants to
                    # set one of the items.  In that situation,
                    # there is only one level in the _call_stack, and we
                    # need to iterate over it here (so we do not allow
                    # the outer portion of this loop to handle the
                    # iteration).  This is indicated by setting the
                    # _iter_stack value to _NotIterable.
                    if self._iter_stack[idx] is _NotIterable:
                        _iter = _slice_generator(
                            *_call[1],
                            iter_over_index=self._iter_over_index,
                            sort=self._sort,
                        )
                        while True:
                            # This ends when the _slice_generator raises
                            # a StopIteration exception
                            self.advance_iter(_iter)
                            # Check to make sure the custom iterator
                            # (i.e._fill_in_known_wildcards) is complete
                            self.advance_iter.check_complete()
                            _comp[_iter.last_index] = _call[2]

                    # The problem here is that _call[1] may be a slice.
                    # If it is, but we are in something like a
                    # _ReferenceDict, where the caller actually wants a
                    # specific index from the slice, we cannot simply
                    # set every element of the slice.  Instead, we will
                    # look for the component (generating a slice if
                    # appropriate).  If it returns a slice, we can use
                    # our current advance_iter to walk it and set only
                    # the appropriate keys
                    try:
                        _tmp = _comp.__getitem__(_call[1])
                    except KeyError:
                        # Since we are slicing, we may only be
                        # interested in things that match.  We will
                        # allow users to (silently) ignore any key
                        # errors generated by concrete indices in the
                        # slice hierarchy...
                        if (
                            self._slice.key_errors_generate_exceptions
                            and not self._iter_over_index
                        ):
                            raise
                        break
                    if _tmp.__class__ is IndexedComponent_slice:
                        # Extract the _slice_generator and evaluate it.
                        assert _tmp._len == 1
                        _iter = _IndexedComponent_slice_iter(
                            _tmp, self.advance_iter, sort=self._sort
                        )
                        for _ in _iter:
                            # Check to make sure the custom iterator
                            # (i.e._fill_in_known_wildcards) is complete
                            self.advance_iter.check_complete()
                            _comp[_iter.get_last_index()] = _call[2]
                        break
                    else:
                        # Check to make sure the custom iterator
                        # (i.e._fill_in_known_wildcards) is complete
                        self.advance_iter.check_complete()
                        # No try-catch, since we know this key is valid
                        _comp[_call[1]] = _call[2]
                    # If we want to support vectorized set_item, e.g.
                    # `m.b[:].v[1] = m.b[:].v[0]`,
                    # we need to cache an iterator over _call[2].
                elif _call[0] == IndexedComponent_slice.del_item:
                    assert idx == self._slice._len - 1
                    # The problem here is that _call[1] may be a slice.
                    # If it is, but we are in something like a
                    # _ReferenceDict, where the caller actually wants a
                    # specific index from the slice, we cannot simply
                    # delete the slice from the component.  Instead, we
                    # will look for the component (generating a slice if
                    # appropriate).  If it returns a slice, we can use
                    # our current advance_iter to walk it and delete the
                    # appropriate keys
                    try:
                        _tmp = _comp.__getitem__(_call[1])
                    except KeyError:
                        # Since we are slicing, we may only be
                        # interested in things that match.  We will
                        # allow users to (silently) ignore any key
                        # errors generated by concrete indices in the
                        # slice hierarchy...
                        if self._slice.key_errors_generate_exceptions:
                            raise
                        break
                    if _tmp.__class__ is IndexedComponent_slice:
                        # Extract the _slice_generator and evaluate it.
                        assert _tmp._len == 1
                        _iter = _IndexedComponent_slice_iter(
                            _tmp, self.advance_iter, sort=self._sort
                        )
                        _idx_to_del = []
                        # Two passes, so that we don't edit the _data
                        # dicts while we are iterating over them
                        for _ in _iter:
                            _idx_to_del.append(_iter.get_last_index())
                        # Check to make sure the custom iterator
                        # (i.e._fill_in_known_wildcards) is complete
                        self.advance_iter.check_complete()
                        for _idx in _idx_to_del:
                            del _comp[_idx]
                        break
                    else:
                        # No try-catch, since we know this key is valid
                        del _comp[_call[1]]
                elif _call[0] == IndexedComponent_slice.del_attribute:
                    assert idx == self._slice._len - 1
                    try:
                        _comp = delattr(_comp, _call[1])
                    except AttributeError:
                        # Since we are slicing, we may only be interested in
                        # things that match.  We will allow users to
                        # (silently) ignore any attribute errors generated
                        # by concrete indices in the slice hierarchy...
                        if self._slice.attribute_errors_generate_exceptions:
                            raise
                        break
                else:
                    raise DeveloperError(
                        "Unexpected entry in IndexedComponent_slice "
                        "_call_stack: %s" % (_call[0],)
                    )
                idx += 1

            if idx == self._slice._len:
                # Check to make sure the custom iterator
                # (i.e._fill_in_known_wildcards) is complete
                self.advance_iter.check_complete()
                # We have a concrete object at the end of the chain. Return it
                return _comp

    def get_last_index(self):
        ans = sum((x.last_index for x in self._iter_stack if x is not None), ())
        if len(ans) == 1:
            return ans[0]
        else:
            return ans

    def get_last_index_wildcards(self):
        """Get a tuple of the values in the wildcard positions for the most
        recent indices corresponding to the last component returned by
        each _slice_generator in the iter stack.

        """
        # This method is how we iterate over keys.
        #
        # last_index is the index corresponding to the most recent
        # component data returned by the corresponding _slice_generator.
        # Extract the indices corresponding to the wildcard positions
        # for that slice.
        ans = sum(
            (
                tuple(
                    x.last_index[i]
                    for i in range(len(x.last_index))
                    if i not in x.fixed
                )
                for x in self._iter_stack
                if x is not None
            ),
            (),
        )
        if not ans:
            return UnindexedComponent_index
        if len(ans) == 1:
            return ans[0]
        else:
            return ans



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


class _robust_sort_keyfcn:
    """Class for robustly generating sortable keys for arbitrary data.

    Generates keys (for use with Python `sorted()` that are
    (str(type_name), val), where val is the actual value (if the type
    is comparable), otherwise the string representation of the value.
    If str() also fails, we fall back on id().

    This allows sorting lists with mixed types in Python3

    We implement this as a callable object so that we can store the
    user's original key function, if provided

    """

    _typemap = {
        int: (1, float.__name__),
        float: (1, float.__name__),
        str: (1, str.__name__),
        tuple: (4, tuple.__name__),
    }

    def __init__(self, key=None):
        self._key = key

    def __call__(self, val):
        """Generate a tuple ( str(type_name), val ) for sorting the value.

        `key=` expects a function.  We are generating a functor so we
        have a convenient place to store the user-provided key and the
        (singleton) _typemap, which maps types to the type-specific
        functions for converting a value to the second argument of the
        sort key.

        """
        if self._key is not None:
            val = self._key(val)

        return self._generate_sort_key(val)

    def _classify_type(self, val):
        _type = val.__class__
        _typename = _type.__name__
        try:
            # 1: Check if the type is comparable.  In Python 3, sorted()
            #    uses "<" to compare objects.
            val < val
            i = 1
            # 1a: Check if the value is comparable to a float.  If
            # it is, sort it as if it were a float.
            try:
                # Extra check that the comparison returns a meaningful result
                if bool(val < 1.0) != bool(1.0 < val or 1.0 == val):
                    _typename = float.__name__
            except:
                pass
        except:
            try:
                # 2: try converting the value to string
                str(val)
                i = 2
            except:
                # 3: fallback on id().  Not deterministic
                #    (run-to-run), but at least is consistent within
                #    this run.
                i = 3
        self._typemap[_type] = i, _typename

    def _generate_sort_key(self, val):
        if val.__class__ not in self._typemap:
            # If this is not a type we have seen before, determine what
            # to use for the second value in the sorting tuple.
            self._classify_type(val)
        i, _typename = self._typemap[val.__class__]
        if i == 1:
            # value type is directly comparable
            return _typename, val
        elif i == 4:
            # nested tuple: recurse into it (so that the tuple is comparable)
            return _typename, tuple(self._generate_sort_key(v) for v in val)
        elif i == 2:
            # value type is convertible to string
            return _typename, str(val)
        else:
            # everything else (including i==3), fall back on id()
            return _typename, id(val)


def sorted_robust(iterable, key=None, reverse=False):
    """Utility to sort an arbitrary iterable.

    This returns the sorted(arg) in a consistent order by first trying
    the standard sort() function, and if that fails (for example with
    mixed-type Sets in Python3), use the _robust_sort_keyfcn utility
    (above) to generate sortable keys.

    Parameters
    ----------
    iterable: iterable
        the source of items to sort
    key: function
        a function of one argument that is used to extract the
        comparison key from each element in `iterable`
    reverse: bool
        if True, the iterable is sorted as if each comparison was reversed.

    Returns
    -------
    list
    """
    # Because we implement this as a "try a normal (fast) sort, then
    # fall back on our slow, but robust sort", we will need to cache the
    # incoming arg: it may be a generator, in which case we would need
    # to exhaust it so we can cache all the values for the case that the
    # first sort attempt fails.  Given that, it is simpler / easier to
    # take *all* incoming args and create a new list, then use list's
    # in-place sort().  By copying *all* incoming data, we avoid
    # possible side effects in the case that the user provided a list.
    ans = list(iterable)
    try:
        ans.sort(key=key, reverse=reverse)
    except:
        ans.sort(key=_robust_sort_keyfcn(key), reverse=reverse)
    return ans



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import collections.abc
import pickle

import pyomo.common.unittest as unittest
from pyomo.common.collections import ComponentSet
from pyomo.core.kernel.variable import variable, variable_dict, variable_list
from pyomo.core.kernel.constraint import constraint, constraint_dict, constraint_list
from pyomo.core.kernel.objective import objective, objective_dict, objective_list
from pyomo.core.kernel.expression import expression, expression_dict, expression_list
from pyomo.core.kernel.block import block, block_dict, block_list
from pyomo.core.kernel.suffix import suffix


class TestComponentSet(unittest.TestCase):
    _components = [
        variable(),
        variable_dict(),
        variable_list(),
        constraint(),
        constraint_dict(),
        constraint_list(),
        objective(),
        objective_dict(),
        objective_list(),
        expression(),
        expression_dict(),
        expression_list(),
        block(),
        block_dict(),
        block_list(),
        suffix(),
    ]

    def test_pickle(self):
        c = ComponentSet()
        self.assertEqual(len(c), 0)
        cup = pickle.loads(pickle.dumps(c))
        self.assertIsNot(cup, c)
        self.assertEqual(len(cup), 0)

        v = variable()
        c.add(v)
        self.assertEqual(len(c), 1)
        self.assertTrue(v in c)
        cup = pickle.loads(pickle.dumps(c))
        vup = cup.pop()
        cup.add(vup)
        self.assertIsNot(cup, c)
        self.assertIsNot(vup, v)
        self.assertEqual(len(cup), 1)
        self.assertTrue(vup in cup)
        self.assertEqual(vup.parent, None)

        b = block()
        V = b.V = variable_list()
        b.V.append(v)
        b.c = c
        self.assertEqual(len(c), 1)
        self.assertTrue(v in c)
        self.assertIs(v.parent, b.V)
        self.assertIs(V.parent, b)
        self.assertIs(b.parent, None)
        bup = pickle.loads(pickle.dumps(b))
        Vup = bup.V
        vup = Vup[0]
        cup = bup.c
        self.assertIsNot(cup, c)
        self.assertIsNot(vup, v)
        self.assertIsNot(Vup, V)
        self.assertIsNot(bup, b)
        self.assertEqual(len(cup), 1)
        self.assertTrue(vup in cup)
        self.assertIs(vup.parent, Vup)
        self.assertIs(Vup.parent, bup)
        self.assertIs(bup.parent, None)

        self.assertEqual(len(c), 1)
        self.assertTrue(v in c)

    def test_init(self):
        cset = ComponentSet()
        cset = ComponentSet(self._components)
        with self.assertRaises(TypeError):
            cset = ComponentSet(*self._components)

    def test_type(self):
        cset = ComponentSet()
        self.assertTrue(isinstance(cset, collections.abc.Set))
        self.assertTrue(isinstance(cset, collections.abc.MutableSet))
        self.assertTrue(issubclass(type(cset), collections.abc.Set))
        self.assertTrue(issubclass(type(cset), collections.abc.MutableSet))

    def test_str(self):
        cset = ComponentSet()
        self.assertEqual(str(cset), "ComponentSet([])")
        cset.update(self._components)
        str(cset)

    def test_len(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        cset.update(self._components)
        self.assertEqual(len(cset), len(self._components))
        cset = ComponentSet(self._components)
        self.assertEqual(len(cset), len(self._components))
        self.assertTrue(len(self._components) > 0)

    def test_iter(self):
        cset = ComponentSet()
        self.assertEqual(list(iter(cset)), [])
        cset.update(self._components)
        ids_seen = set()
        for c in cset:
            ids_seen.add(id(c))
        self.assertEqual(ids_seen, set(id(c) for c in self._components))

    def set_add(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        for i, c in enumerate(self._components):
            self.assertTrue(c not in cset)
            cset.add(c)
            self.assertTrue(c in cset)
            self.assertEqual(len(cset), i + 1)
        self.assertEqual(len(cset), len(self._components))
        for c in self._components:
            self.assertTrue(c in cset)
            cset.add(c)
            self.assertTrue(c in cset)
            self.assertEqual(len(cset), len(self._components))

    def test_pop(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        with self.assertRaises(KeyError):
            cset.pop()
        v = variable()
        cset.add(v)
        self.assertTrue(v in cset)
        self.assertEqual(len(cset), 1)
        v_ = cset.pop()
        self.assertIs(v, v_)
        self.assertTrue(v not in cset)
        self.assertEqual(len(cset), 0)

    def test_update(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        cset.update(self._components)
        self.assertEqual(len(cset), len(self._components))
        for c in self._components:
            self.assertTrue(c in cset)

    def test_clear(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        cset.update(self._components)
        self.assertEqual(len(cset), len(self._components))
        cset.clear()
        self.assertEqual(len(cset), 0)

    def test_remove(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        cset.update(self._components)
        self.assertEqual(len(cset), len(self._components))
        for i, c in enumerate(self._components):
            cset.remove(c)
            self.assertEqual(len(cset), len(self._components) - (i + 1))
        for c in self._components:
            self.assertTrue(c not in cset)
            with self.assertRaises(KeyError):
                cset.remove(c)

    def test_discard(self):
        cset = ComponentSet()
        self.assertEqual(len(cset), 0)
        cset.update(self._components)
        self.assertEqual(len(cset), len(self._components))
        for i, c in enumerate(self._components):
            cset.discard(c)
            self.assertEqual(len(cset), len(self._components) - (i + 1))
        for c in self._components:
            self.assertTrue(c not in cset)
            cset.discard(c)

    def test_isdisjoint(self):
        cset1 = ComponentSet()
        cset2 = ComponentSet()
        self.assertTrue(cset1.isdisjoint(cset2))
        self.assertTrue(cset2.isdisjoint(cset1))
        v = variable()
        cset1.add(v)
        self.assertTrue(cset1.isdisjoint(cset2))
        self.assertTrue(cset2.isdisjoint(cset1))
        cset2.add(v)
        self.assertFalse(cset1.isdisjoint(cset2))
        self.assertFalse(cset2.isdisjoint(cset1))

    def test_misc_set_ops(self):
        v1 = variable()
        cset1 = ComponentSet([v1])
        v2 = variable()
        cset2 = ComponentSet([v2])
        cset3 = ComponentSet([v1, v2])
        empty = ComponentSet([])
        self.assertEqual(cset1 | cset2, cset3)
        self.assertEqual((cset1 | cset2) - cset3, empty)
        self.assertEqual(cset1 ^ cset2, cset3)
        self.assertEqual(cset1 ^ cset3, cset2)
        self.assertEqual(cset2 ^ cset3, cset1)
        self.assertEqual(cset1 & cset2, empty)
        self.assertEqual(cset1 & cset3, cset1)
        self.assertEqual(cset2 & cset3, cset2)

    def test_eq(self):
        cset1 = ComponentSet()
        self.assertEqual(cset1, set())
        self.assertTrue(cset1 == set())
        self.assertNotEqual(cset1, list())
        self.assertFalse(cset1 == list())
        self.assertNotEqual(cset1, tuple())
        self.assertFalse(cset1 == tuple())
        self.assertNotEqual(cset1, dict())
        self.assertFalse(cset1 == dict())

        cset1.update(self._components)
        self.assertNotEqual(cset1, set())
        self.assertFalse(cset1 == set())
        self.assertNotEqual(cset1, list())
        self.assertFalse(cset1 == list())
        self.assertNotEqual(cset1, tuple())
        self.assertFalse(cset1 == tuple())
        self.assertNotEqual(cset1, dict())
        self.assertFalse(cset1 == dict())

        self.assertTrue(cset1 == cset1)
        self.assertEqual(cset1, cset1)

        cset2 = ComponentSet(self._components)
        self.assertTrue(cset2 == cset1)
        self.assertFalse(cset2 != cset1)
        self.assertEqual(cset2, cset1)
        self.assertTrue(cset1 == cset2)
        self.assertFalse(cset1 != cset2)
        self.assertEqual(cset1, cset2)

        cset2.remove(self._components[0])
        self.assertFalse(cset2 == cset1)
        self.assertTrue(cset2 != cset1)
        self.assertNotEqual(cset2, cset1)
        self.assertFalse(cset1 == cset2)
        self.assertTrue(cset1 != cset2)
        self.assertNotEqual(cset1, cset2)

        cset2.add(variable())
        self.assertFalse(cset2 == cset1)
        self.assertTrue(cset2 != cset1)
        self.assertNotEqual(cset2, cset1)
        self.assertFalse(cset1 == cset2)
        self.assertTrue(cset1 != cset2)
        self.assertNotEqual(cset1, cset2)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from io import StringIO
import logging

from pyomo.common.dependencies import attempt_import, scipy_available, numpy_available
from pyomo.common.log import LoggingIntercept
import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise import PiecewiseLinearFunction
from pyomo.contrib.piecewise.transform.nonlinear_to_pwl import (
    NonlinearToPWL,
    DomainPartitioningMethod,
    lineartree_available,
    sklearn_available,
)
from pyomo.core.base.expression import _ExpressionData
from pyomo.core.expr.compare import (
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)
from pyomo.core.expr.numeric_expr import SumExpression
from pyomo.environ import (
    Binary,
    ConcreteModel,
    Var,
    Constraint,
    Integers,
    TransformationFactory,
    log,
    Objective,
    Reals,
    SolverFactory,
    TerminationCondition,
    value,
)

gurobi_available = (
    SolverFactory('gurobi').available(exception_flag=False)
    and SolverFactory('gurobi').license_is_valid()
)


class TestNonlinearToPWL_1D(unittest.TestCase):
    def make_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(1, 10))
        m.cons = Constraint(expr=log(m.x) >= 0.35)

        return m

    def check_pw_linear_log_x(self, m, pwlf, x1, x2, x3):
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')

        points = [(x1,), (x2,), (x3,)]
        self.assertEqual(pwlf._simplices, [(0, 1), (1, 2)])
        self.assertEqual(pwlf._points, points)
        self.assertEqual(len(pwlf._linear_functions), 2)

        assertExpressionsStructurallyEqual(
            self,
            pwlf._linear_functions[0](m.x),
            ((log(x2) - log(x1)) / (x2 - x1)) * m.x
            + (log(x2) - ((log(x2) - log(x1)) / (x2 - x1)) * x2),
            places=7,
        )
        assertExpressionsStructurallyEqual(
            self,
            pwlf._linear_functions[1](m.x),
            ((log(x3) - log(x2)) / (x3 - x2)) * m.x
            + (log(x3) - ((log(x3) - log(x2)) / (x3 - x2)) * x3),
            places=7,
        )

        self.assertEqual(len(pwlf._expressions), 1)
        new_cons = n_to_pwl.get_transformed_component(m.cons)
        self.assertTrue(new_cons.active)
        self.assertIs(
            new_cons.body, pwlf._expressions[pwlf._expression_ids[new_cons.body.expr]]
        )
        self.assertIsNone(new_cons.ub)
        self.assertEqual(new_cons.lb, 0.35)
        self.assertIs(n_to_pwl.get_src_component(new_cons), m.cons)

        quadratic = n_to_pwl.get_transformed_quadratic_constraints(m)
        self.assertEqual(len(quadratic), 0)
        nonlinear = n_to_pwl.get_transformed_nonlinear_constraints(m)
        self.assertEqual(len(nonlinear), 1)
        self.assertIn(m.cons, nonlinear)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_log_constraint_uniform_grid(self):
        m = self.make_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=3,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
        )

        # cons is transformed
        self.assertFalse(m.cons.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        points = [(1.0009,), (5.5,), (9.9991,)]
        x1, x2, x3 = 1.0009, 5.5, 9.9991
        self.check_pw_linear_log_x(m, pwlf, x1, x2, x3)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_clone_transformed_model(self):
        m = self.make_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=3,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
        )

        twin = m.clone()

        # cons is transformed
        self.assertFalse(twin.cons.active)

        pwlf = list(
            twin.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        points = [(1.0009,), (5.5,), (9.9991,)]
        x1, x2, x3 = 1.0009, 5.5, 9.9991

        self.check_pw_linear_log_x(twin, pwlf, x1, x2, x3)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_log_constraint_random_grid(self):
        m = self.make_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        # [ESJ 3/30/24]: The seed is actually set in the function for getting
        # the points right now, so this will be deterministic.
        n_to_pwl.apply_to(
            m,
            num_points=3,
            domain_partitioning_method=DomainPartitioningMethod.RANDOM_GRID,
        )

        # cons is transformed
        self.assertFalse(m.cons.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        x1 = 4.370861069626263
        x2 = 7.587945476302646
        x3 = 9.556428757689245
        self.check_pw_linear_log_x(m, pwlf, x1, x2, x3)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_do_not_transform_quadratic_constraint(self):
        m = self.make_model()
        m.quad = Constraint(expr=m.x**2 <= 9)
        m.lin = Constraint(expr=m.x >= 2)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=3,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            approximate_quadratic_constraints=False,
        )

        # cons is transformed
        self.assertFalse(m.cons.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        points = [(1.0009,), (5.5,), (9.9991,)]
        x1, x2, x3 = 1.0009, 5.5, 9.9991
        self.check_pw_linear_log_x(m, pwlf, x1, x2, x3)

        # quad is not
        self.assertTrue(m.quad.active)
        # neither is the linear one
        self.assertTrue(m.lin.active)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_constraint_target(self):
        m = self.make_model()
        m.quad = Constraint(expr=m.x**2 <= 9)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=3,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            targets=[m.cons],
        )

        # cons is transformed
        self.assertFalse(m.cons.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        points = [(1.0009,), (5.5,), (9.9991,)]
        x1, x2, x3 = 1.0009, 5.5, 9.9991
        self.check_pw_linear_log_x(m, pwlf, x1, x2, x3)

        # quad is not
        self.assertTrue(m.quad.active)

    def test_crazy_target_error(self):
        m = self.make_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        with self.assertRaisesRegex(
            ValueError,
            "Target 'x' is not a Block, Constraint, or Objective. It "
            "is of type '<class 'pyomo.core.base.var.ScalarVar'>' and cannot "
            "be transformed.",
        ):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
                targets=[m.x],
            )

    def test_cannot_approximate_constraints_with_unbounded_vars(self):
        m = ConcreteModel()
        m.x = Var()
        m.quad = Constraint(expr=m.x**2 <= 9)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        with self.assertRaisesRegex(
            ValueError,
            "Cannot automatically approximate constraints with unbounded "
            "variables. Var 'x' appearing in component 'quad' is missing "
            "at least one bound",
        ):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            )

    def test_error_for_non_separable_exceeding_max_dimension(self):
        m = ConcreteModel()
        m.x = Var([0, 1, 2, 3, 4], bounds=(-4, 5))
        m.ick = Constraint(expr=m.x[0] ** (m.x[1] * m.x[2] * m.x[3] * m.x[4]) <= 8)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        with self.assertRaisesRegex(
            ValueError,
            "Not approximating expression for component 'ick' as "
            "it exceeds the maximum dimension of 4. Try increasing "
            "'max_dimension' or additively separating the expression.",
        ):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
                max_dimension=4,
            )

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_do_not_additively_decompose_below_min_dimension(self):
        m = ConcreteModel()
        m.x = Var([0, 1, 2, 3, 4], bounds=(-4, 5))
        m.c = Constraint(expr=m.x[0] * m.x[1] + m.x[3] <= 4)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=3,
            additively_decompose=True,
            min_dimension_to_additively_decompose=4,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
        )

        transformed_c = n_to_pwl.get_transformed_component(m.c)
        # This is only approximated by one pwlf:
        self.assertIsInstance(transformed_c.body, _ExpressionData)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_uniform_sampling_discrete_vars(self):
        m = ConcreteModel()
        m.x = Var(['rocky', 'bullwinkle'], domain=Binary)
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.c = Constraint(expr=m.x['rocky'] * m.x['bullwinkle'] + m.y <= 4)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core', logging.WARNING):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                additively_decompose=False,
                domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            )
        # No warnings (this is to check that we aren't emitting a bunch of
        # warnings about setting variables outside of their domains)
        self.assertEqual(output.getvalue().strip(), "")

        transformed_c = n_to_pwl.get_transformed_component(m.c)
        pwlf = transformed_c.body.expr.pw_linear_function

        # should sample 0, 1 for th m.x's
        # should sample 0, 2, 5 for m.y (because of half to even rounding (*sigh*))
        points = set(pwlf._points)
        self.assertEqual(len(points), 12)
        for x in [0, 1]:
            for y in [0, 1]:
                for z in [0, 2, 5]:
                    self.assertIn((x, y, z), points)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_uniform_sampling_discrete_vars(self):
        m = ConcreteModel()
        m.x = Var(['rocky', 'bullwinkle'], domain=Binary)
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.c = Constraint(expr=m.x['rocky'] * m.x['bullwinkle'] + m.y <= 4)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core', logging.WARNING):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                additively_decompose=False,
                domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            )
        # No warnings (this is to check that we aren't emitting a bunch of
        # warnings about setting variables outside of their domains)
        self.assertEqual(output.getvalue().strip(), "")

        transformed_c = n_to_pwl.get_transformed_component(m.c)
        pwlf = transformed_c.body.expr.pw_linear_function

        # should sample 0, 1 for th m.x's
        # should sample 0, 2, 5 for m.y (because of half to even rounding (*sigh*))
        points = set(pwlf._points)
        self.assertEqual(len(points), 12)
        for x in [0, 1]:
            for y in [0, 1]:
                for z in [0, 2, 5]:
                    self.assertIn((x, y, z), points)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_random_sampling_discrete_vars(self):
        m = ConcreteModel()
        m.x = Var(['rocky', 'bullwinkle'], domain=Binary)
        m.y = Var(domain=Integers, bounds=(0, 5))
        m.c = Constraint(expr=m.x['rocky'] * m.x['bullwinkle'] + m.y <= 4)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        output = StringIO()
        with LoggingIntercept(output, 'pyomo.core', logging.WARNING):
            n_to_pwl.apply_to(
                m,
                num_points=3,
                additively_decompose=False,
                domain_partitioning_method=DomainPartitioningMethod.RANDOM_GRID,
            )
        # No warnings (this is to check that we aren't emitting a bunch of
        # warnings about setting variables outside of their domains)
        self.assertEqual(output.getvalue().strip(), "")

        transformed_c = n_to_pwl.get_transformed_component(m.c)
        pwlf = transformed_c.body.expr.pw_linear_function

        # should sample 0, 1 for th m.x's
        # Happen to get 0, 1, 5 for m.y
        points = set(pwlf._points)
        self.assertEqual(len(points), 12)
        for x in [0, 1]:
            for y in [0, 1]:
                for z in [0, 1, 5]:
                    self.assertIn((x, y, z), points)


class TestNonlinearToPWL_2D(unittest.TestCase):
    def make_paraboloid_model(self):
        m = ConcreteModel()
        m.x1 = Var(bounds=(0, 3))
        m.x2 = Var(bounds=(1, 7))
        m.obj = Objective(expr=m.x1**2 + m.x2**2)

        return m

    def check_pw_linear_paraboloid(self, m, pwlf, x1, x2, y1, y2):
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        points = [(x1, y1), (x1, y2), (x2, y1), (x2, y2)]
        self.assertEqual(pwlf._points, points)
        self.assertEqual(pwlf._simplices, [(0, 1, 3), (0, 2, 3)])
        self.assertEqual(len(pwlf._linear_functions), 2)

        # just check that the linear functions make sense--they intersect the
        # paraboloid at the vertices of the simplices.
        self.assertAlmostEqual(pwlf._linear_functions[0](x1, y1), x1**2 + y1**2)
        self.assertAlmostEqual(pwlf._linear_functions[0](x1, y2), x1**2 + y2**2)
        self.assertAlmostEqual(pwlf._linear_functions[0](x2, y2), x2**2 + y2**2)

        self.assertAlmostEqual(pwlf._linear_functions[1](x1, y1), x1**2 + y1**2)
        self.assertAlmostEqual(pwlf._linear_functions[1](x2, y1), x2**2 + y1**2)
        self.assertAlmostEqual(pwlf._linear_functions[1](x2, y2), x2**2 + y2**2)

        self.assertEqual(len(pwlf._expressions), 1)
        new_obj = n_to_pwl.get_transformed_component(m.obj)
        self.assertTrue(new_obj.active)
        self.assertIs(
            new_obj.expr, pwlf._expressions[pwlf._expression_ids[new_obj.expr.expr]]
        )
        self.assertIs(n_to_pwl.get_src_component(new_obj), m.obj)

        quadratic = n_to_pwl.get_transformed_quadratic_constraints(m)
        self.assertEqual(len(quadratic), 0)
        nonlinear = n_to_pwl.get_transformed_nonlinear_constraints(m)
        self.assertEqual(len(nonlinear), 0)
        quadratic = n_to_pwl.get_transformed_quadratic_objectives(m)
        self.assertEqual(len(quadratic), 1)
        self.assertIn(m.obj, quadratic)
        nonlinear = n_to_pwl.get_transformed_nonlinear_objectives(m)
        self.assertEqual(len(nonlinear), 0)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_paraboloid_objective_uniform_grid(self):
        m = self.make_paraboloid_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=2,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
        )

        # check obj is transformed
        self.assertFalse(m.obj.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        x1 = 0.00030000000000000003
        x2 = 2.9997
        y1 = 1.0006
        y2 = 6.9994

        self.check_pw_linear_paraboloid(m, pwlf, x1, x2, y1, y2)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_multivariate_clone(self):
        m = self.make_paraboloid_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=2,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
        )

        twin = m.clone()

        # check obj is transformed
        self.assertFalse(twin.obj.active)

        pwlf = list(
            twin.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        x1 = 0.00030000000000000003
        x2 = 2.9997
        y1 = 1.0006
        y2 = 6.9994

        self.check_pw_linear_paraboloid(twin, pwlf, x1, x2, y1, y2)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_objective_target(self):
        m = self.make_paraboloid_model()

        m.some_other_nonlinear_constraint = Constraint(expr=m.x1**3 + m.x2 <= 6)

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=2,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            targets=[m.obj],
        )

        # check obj is transformed
        self.assertFalse(m.obj.active)

        pwlf = list(
            m.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(pwlf), 1)
        pwlf = pwlf[0]

        x1 = 0.00030000000000000003
        x2 = 2.9997
        y1 = 1.0006
        y2 = 6.9994

        self.check_pw_linear_paraboloid(m, pwlf, x1, x2, y1, y2)

        # and check that the constraint isn't transformed
        self.assertTrue(m.some_other_nonlinear_constraint.active)

    def test_do_not_transform_quadratic_objective(self):
        m = self.make_paraboloid_model()

        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=2,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            approximate_quadratic_objectives=False,
        )

        # check obj is *not* transformed
        self.assertTrue(m.obj.active)

        quadratic = n_to_pwl.get_transformed_quadratic_constraints(m)
        self.assertEqual(len(quadratic), 0)
        nonlinear = n_to_pwl.get_transformed_nonlinear_constraints(m)
        self.assertEqual(len(nonlinear), 0)
        quadratic = n_to_pwl.get_transformed_quadratic_objectives(m)
        self.assertEqual(len(quadratic), 0)
        nonlinear = n_to_pwl.get_transformed_nonlinear_objectives(m)
        self.assertEqual(len(nonlinear), 0)


@unittest.skipUnless(lineartree_available, "lineartree not available")
@unittest.skipUnless(sklearn_available, "sklearn not available")
class TestLinearTreeDomainPartitioning(unittest.TestCase):
    def make_absolute_value_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(-10, 10))
        m.obj = Objective(expr=abs(m.x))

        return m

    def test_linear_model_tree_uniform(self):
        m = self.make_absolute_value_model()
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=301,  # sample a lot so we train a good tree
            domain_partitioning_method=DomainPartitioningMethod.LINEAR_MODEL_TREE_UNIFORM,
            linear_tree_max_depth=1,  # force parsimony
        )

        transformed_obj = n_to_pwl.get_transformed_component(m.obj)
        pwlf = transformed_obj.expr.expr.pw_linear_function

        self.assertEqual(len(pwlf._simplices), 2)
        self.assertEqual(pwlf._simplices, [(0, 1), (1, 2)])
        self.assertEqual(pwlf._points, [(-10,), (-0.08402,), (10,)])
        self.assertEqual(len(pwlf._linear_functions), 2)
        assertExpressionsEqual(self, pwlf._linear_functions[0](m.x), -1.0 * m.x)
        assertExpressionsStructurallyEqual(
            self,
            pwlf._linear_functions[1](m.x),
            # pretty close to m.x, but we're a bit off because we don't have 0
            # as a breakpoint.
            0.9833360108369479 * m.x + 0.16663989163052034,
            places=6,
        )

    def test_linear_model_tree_random(self):
        m = self.make_absolute_value_model()
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=300,  # sample a lot so we train a good tree
            domain_partitioning_method=DomainPartitioningMethod.LINEAR_MODEL_TREE_RANDOM,
            linear_tree_max_depth=1,  # force parsimony
        )

        transformed_obj = n_to_pwl.get_transformed_component(m.obj)
        pwlf = transformed_obj.expr.expr.pw_linear_function

        self.assertEqual(len(pwlf._simplices), 2)
        self.assertEqual(pwlf._simplices, [(0, 1), (1, 2)])
        self.assertEqual(pwlf._points, [(-10,), (-0.03638,), (10,)])
        self.assertEqual(len(pwlf._linear_functions), 2)
        assertExpressionsEqual(self, pwlf._linear_functions[0](m.x), -1.0 * m.x)
        assertExpressionsStructurallyEqual(
            self,
            pwlf._linear_functions[1](m.x),
            # pretty close to m.x, but we're a bit off because we don't have 0
            # as a breakpoint.
            0.9927503741388829 * m.x + 0.07249625861117256,
            places=7,
        )

    def test_linear_model_tree_random_auto_depth_tree(self):
        m = self.make_absolute_value_model()
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        n_to_pwl.apply_to(
            m,
            num_points=100,  # sample a lot but not too many because this one is
            # more prone to overfitting
            domain_partitioning_method=DomainPartitioningMethod.LINEAR_MODEL_TREE_RANDOM,
        )

        transformed_obj = n_to_pwl.get_transformed_component(m.obj)
        pwlf = transformed_obj.expr.expr.pw_linear_function

        print(pwlf._simplices)
        print(pwlf._points)
        for f in pwlf._linear_functions:
            print(f(m.x))

        # We end up with 8, which is just what happens, but it's not a terrible
        # approximation
        self.assertEqual(len(pwlf._simplices), 8)
        self.assertEqual(
            pwlf._simplices,
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8)],
        )
        self.assertStructuredAlmostEqual(
            pwlf._points,
            [
                (-10,),
                (-9.24119,),
                (-8.71428,),
                (-8.11135,),
                (0.06048,),
                (0.70015,),
                (1.9285,),
                (2.15597,),
                (10,),
            ],
            places=4,
        )
        self.assertEqual(len(pwlf._linear_functions), 8)
        for i in range(3):
            assertExpressionsEqual(self, pwlf._linear_functions[i](m.x), -1.0 * m.x)
        assertExpressionsStructurallyEqual(
            self,
            pwlf._linear_functions[3](m.x),
            # pretty close to - m.x, but we're a bit off because we don't have 0
            # as a breakpoint.
            -0.9851979299618323 * m.x + 0.12006477080409184,
            places=6,
        )
        for i in range(4, 8):
            assertExpressionsEqual(self, pwlf._linear_functions[i](m.x), m.x)


class TestNonlinearToPWLIntegration(unittest.TestCase):
    @unittest.skipUnless(gurobi_available, "Gurobi is not available")
    @unittest.skipUnless(scipy_available, "Scipy is not available")
    def test_transform_and_solve_additively_decomposes_model(self):
        # A bit of an integration test to make sure that we build additively
        # decomposed pw-linear approximations in such a way that they are
        # transformed to MILP and solved correctly. (Largely because we have to
        # be careful to make sure that we don't ever directly insert
        # PiecewiseLinearExpression objects into expressions and are instead
        # using the ExpressionData that points to them (and will eventually be
        # replaced in transformation))
        m = ConcreteModel()
        m.x1 = Var(within=Reals, bounds=(0, 2), initialize=1.745)
        m.x4 = Var(within=Reals, bounds=(0, 5), initialize=3.048)
        m.x7 = Var(within=Reals, bounds=(0.9, 0.95), initialize=0.928)
        m.obj = Objective(expr=-6.3 * m.x4 * m.x7 + 5.04 * m.x1)
        n_to_pwl = TransformationFactory('contrib.piecewise.nonlinear_to_pwl')
        xm = n_to_pwl.create_using(
            m,
            num_points=4,
            domain_partitioning_method=DomainPartitioningMethod.UNIFORM_GRID,
            additively_decompose=True,
        )

        self.assertFalse(xm.obj.active)
        new_obj = n_to_pwl.get_transformed_component(xm.obj)
        self.assertIs(n_to_pwl.get_src_component(new_obj), xm.obj)
        self.assertTrue(new_obj.active)
        # two terms
        self.assertIsInstance(new_obj.expr, SumExpression)
        self.assertEqual(len(new_obj.expr.args), 2)
        first = new_obj.expr.args[0]
        pwlf = first.expr.pw_linear_function
        all_pwlf = list(
            xm.component_data_objects(PiecewiseLinearFunction, descend_into=True)
        )
        self.assertEqual(len(all_pwlf), 1)
        # It is on the active tree.
        self.assertIs(pwlf, all_pwlf[0])

        second = new_obj.expr.args[1]
        assertExpressionsEqual(self, second, 5.04 * xm.x1)

        objs = n_to_pwl.get_transformed_nonlinear_objectives(xm)
        self.assertEqual(len(objs), 0)
        objs = n_to_pwl.get_transformed_quadratic_objectives(xm)
        self.assertEqual(len(objs), 1)
        self.assertIn(xm.obj, objs)
        self.assertEqual(len(n_to_pwl.get_transformed_nonlinear_constraints(xm)), 0)
        self.assertEqual(len(n_to_pwl.get_transformed_quadratic_constraints(xm)), 0)

        TransformationFactory('contrib.piecewise.outer_repn_gdp').apply_to(xm)
        TransformationFactory('gdp.bigm').apply_to(xm)
        opt = SolverFactory('gurobi')
        results = opt.solve(xm)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )

        # solve the original
        opt.options['NonConvex'] = 2
        results = opt.solve(m)
        self.assertEqual(
            results.solver.termination_condition, TerminationCondition.optimal
        )

        # Not a bad approximation:
        self.assertAlmostEqual(value(xm.obj), value(m.obj), places=2)

        self.assertAlmostEqual(value(xm.x4), value(m.x4), places=3)
        self.assertAlmostEqual(value(xm.x7), value(m.x7), places=4)
        self.assertAlmostEqual(value(xm.x1), value(m.x1), places=7)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
import pyomo.common.unittest as unittest
from pyomo.contrib.appsi.base import TerminationCondition, Results, PersistentSolver
from pyomo.contrib.appsi.solvers.wntr import Wntr, wntr_available
import math

_default_wntr_options = dict(TOL=1e-8)


@unittest.skipUnless(wntr_available, 'wntr is not available')
class TestWntrPersistent(unittest.TestCase):
    def test_param_updates(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.p = pyo.Param(initialize=1, mutable=True)
        m.c = pyo.Constraint(expr=m.x == m.p)
        opt = Wntr()
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 1)

        m.p.value = 2
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 2)

    def test_remove_add_constraint(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.y == (m.x - 1) ** 2)
        m.c2 = pyo.Constraint(expr=m.y == pyo.exp(m.x))
        opt = Wntr()
        opt.config.symbolic_solver_labels = True
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 1)

        del m.c2
        m.c2 = pyo.Constraint(expr=m.y == pyo.log(m.x))
        m.x.value = 0.5
        m.y.value = 0.5
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 0)

    def test_fixed_var(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.y == (m.x - 1) ** 2)
        m.x.fix(0.5)
        opt = Wntr()
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.5)
        self.assertAlmostEqual(m.y.value, 0.25)

        m.x.unfix()
        m.c2 = pyo.Constraint(expr=m.y == pyo.exp(m.x))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0)
        self.assertAlmostEqual(m.y.value, 1)

        m.x.fix(0.5)
        del m.c2
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.5)
        self.assertAlmostEqual(m.y.value, 0.25)

    def test_remove_variables_params(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.z = pyo.Var()
        m.z.fix(0)
        m.px = pyo.Param(mutable=True, initialize=1)
        m.py = pyo.Param(mutable=True, initialize=1)
        m.c1 = pyo.Constraint(expr=m.x == m.px)
        m.c2 = pyo.Constraint(expr=m.y == m.py)
        opt = Wntr()
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 1)
        self.assertAlmostEqual(m.y.value, 1)
        self.assertAlmostEqual(m.z.value, 0)

        del m.c2
        del m.y
        del m.py
        m.z.value = 2
        m.px.value = 2
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 2)
        self.assertAlmostEqual(m.z.value, 2)

        del m.z
        m.px.value = 3
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 3)

    def test_get_primals(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        m.c1 = pyo.Constraint(expr=m.y == (m.x - 1) ** 2)
        m.c2 = pyo.Constraint(expr=m.y == pyo.exp(m.x))
        opt = Wntr()
        opt.config.load_solution = False
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, None)
        self.assertAlmostEqual(m.y.value, None)
        primals = opt.get_primals()
        self.assertAlmostEqual(primals[m.x], 0)
        self.assertAlmostEqual(primals[m.y], 1)

    def test_operators(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var(initialize=1)
        m.c1 = pyo.Constraint(expr=2 / m.x == 1)
        opt = Wntr()
        opt.wntr_options.update(_default_wntr_options)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 2)

        del m.c1
        m.x.value = 0
        m.c1 = pyo.Constraint(expr=pyo.sin(m.x) == math.sin(math.pi / 4))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, math.pi / 4)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.cos(m.x) == 0)
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, math.pi / 2)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.tan(m.x) == 1)
        m.x.value = 0
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, math.pi / 4)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.asin(m.x) == math.asin(0.5))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.5)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.acos(m.x) == math.acos(0.6))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.6)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.atan(m.x) == math.atan(0.5))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.5)

        del m.c1
        m.c1 = pyo.Constraint(expr=pyo.sqrt(m.x) == math.sqrt(0.6))
        res = opt.solve(m)
        self.assertEqual(res.termination_condition, TerminationCondition.optimal)
        self.assertAlmostEqual(m.x.value, 0.6)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import os
import subprocess
import sys

import pyomo.environ as pyo

from pyomo.common import unittest, Executable
from pyomo.common.config import ConfigDict
from pyomo.common.errors import ApplicationError
from pyomo.common.fileutils import ExecutableData
from pyomo.common.log import LoggingIntercept
from pyomo.common.tempfiles import TempfileManager
from pyomo.core.base import SymbolMap
from pyomo.core.base.label import NumericLabeler
from pyomo.opt.base import SolverFactory
from pyomo.repn.plugins.gams_writer_v2 import GAMSWriter
from pyomo.contrib.solver.common.util import (
    NoDualsError,
    NoOptimalSolutionError,
    NoReducedCostsError,
    NoSolutionError,
)
from pyomo.contrib.solver.common.base import Availability
from pyomo.contrib.solver.common.results import TerminationCondition, SolutionStatus
import pyomo.contrib.solver.solvers.gams as gams

"""
Formatted after pyomo/pyomo/contrib/solver/test/solvers/test_ipopt.py
"""


gams_available = gams.GAMS().available()


@unittest.skipIf(not gams_available, "The 'gams' command is not available")
@unittest.pytest.mark.solver("gams")
class TestGAMSSolverConfig(unittest.TestCase):
    def test_default_instantiation(self):
        config = gams.GAMSConfig()
        # Should be inherited
        self.assertIsNone(config._description)
        self.assertEqual(config._visibility, 0)
        self.assertFalse(config.tee)
        self.assertTrue(config.load_solutions)
        self.assertTrue(config.raise_exception_on_nonoptimal_result)
        self.assertFalse(config.symbolic_solver_labels)
        self.assertIsNone(config.timer)
        self.assertIsNone(config.threads)
        self.assertIsNone(config.time_limit)
        # Unique to this object
        self.assertIsInstance(config.executable, type(Executable('path')))
        self.assertIsInstance(config.writer_config, type(GAMSWriter.CONFIG()))

    def test_custom_instantiation(self):
        config = gams.GAMSConfig(description="A description")
        config.tee = True
        self.assertTrue(config.tee)
        self.assertEqual(config._description, "A description")
        self.assertIsNone(config.time_limit)
        # Default should be `gams`
        self.assertIsNotNone(str(config.executable))
        self.assertIn('gams', str(config.executable))
        # Set to a totally bogus path
        config.executable = Executable('/bogus/path')
        self.assertIsNone(config.executable.executable)
        self.assertFalse(config.executable.available())


@unittest.pytest.mark.solver("gams")
class TestGAMSSolutionLoader(unittest.TestCase):
    def test_get_reduced_costs_error(self):
        loader = gams.GMSSolutionLoader(None, None)
        with self.assertRaises(NoSolutionError):
            loader.get_primals()
        with self.assertRaises(NoDualsError):
            loader.get_duals()
        with self.assertRaises(NoReducedCostsError):
            loader.get_reduced_costs()

        # Set _gms_info to something completely bogus but is not None
        # Set the var_symbol_map and con_symbol_map to empty SymbolMap object type
        class GAMSInfo:
            pass

        class GDXData:
            pass

        loader._gms_info = GAMSInfo()
        loader._gms_info.var_symbol_map = SymbolMap(NumericLabeler('x'))
        loader._gms_info.con_symbol_map = SymbolMap(NumericLabeler('c'))

        # We are asserting if there is no solution, the SymbolMap for
        # variable length must be 0
        loader.get_primals()

        # if the model is infeasible, no dual information is returned
        with self.assertRaises(NoDualsError):
            loader.get_duals()


@unittest.skipIf(not gams_available, "The 'gams' command is not available")
@unittest.pytest.mark.solver("gams")
class TestGAMSInterface(unittest.TestCase):
    # _simple_model and _run_simple_model are standalone functions to
    # test gams execution
    def _simple_model(self, n):
        return """
            option limrow = 0;
            option limcol = 0;
            option solprint = off;
            set I / 1 * %s /;
            variables ans;
            positive variables x(I);
            equations obj;
            obj.. ans =g= sum(I, x(I));
            model test / all /;
            solve test using lp minimizing ans;
            """ % (n,)

    def _run_simple_model(self, config, n):
        solver_exec = config.executable.path()
        if solver_exec is None:
            return False
        with TempfileManager.new_context() as tempfile:
            tmpdir = tempfile.mkdtemp()
            test = os.path.join(tmpdir, 'test.gms')
            with open(test, 'w') as FILE:
                FILE.write(self._simple_model(n))
            result = subprocess.run(
                [solver_exec, test, "curdir=" + tmpdir, 'lo=0'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            return not result.returncode

    def test_class_member_list(self):
        opt = gams.GAMS()
        expected_list = [
            'CONFIG',
            'available',
            'config',
            'api_version',
            'is_persistent',
            'name',
            'solve',
            'version',
            # 'license_is_valid', # DEPRECATED
        ]
        method_list = [method for method in dir(opt) if method.startswith('_') is False]
        self.assertEqual(sorted(expected_list), sorted(method_list))

    def test_default_instantiation(self):
        opt = gams.GAMS()
        self.assertFalse(opt.is_persistent())
        self.assertTrue(opt.version())
        self.assertEqual(opt.name, 'gams')
        self.assertEqual(opt.CONFIG, opt.config)
        self.assertTrue(opt.available())

    def test_context_manager(self):
        with gams.GAMS() as opt:
            self.assertFalse(opt.is_persistent())
            self.assertIsNotNone(opt.version())
            self.assertEqual(opt.name, 'gams')
            self.assertEqual(opt.CONFIG, opt.config)
            self.assertTrue(opt.available())

    def test_available_cache(self):
        opt = gams.GAMS()
        self.assertTrue(
            opt.available() in {Availability.FullLicense, Availability.LimitedLicense}
        )
        # Now we will try with a custom config that has a fake path
        config = gams.GAMSConfig()
        config.executable = Executable('/a/bogus/path')
        opt = gams.GAMS()
        opt.config = config
        self.assertTrue(opt.available() == Availability.NotFound)

    def test_version(self):
        # GAMS should be available...
        opt = gams.GAMS()
        ver = opt.version()
        self.assertIsInstance(ver, tuple)
        self.assertEqual(len(ver), 3)
        self.assertIsInstance(ver[0], int)
        self.assertIsInstance(ver[1], int)
        self.assertIsInstance(ver[2], int)
        self.assertGreater(ver[0], 0)
        self.assertGreaterEqual(ver[1], 0)
        self.assertGreaterEqual(ver[2], 0)

        # Now we will try with a custom config that has a fake path
        config = gams.GAMSConfig()
        config.executable = Executable('/a/bogus/path')
        opt = gams.GAMS()
        opt.config = config
        self.assertIsNone(opt.version())

        # Now try pointing to "executables" that are not GAMS
        with TempfileManager as tmp:
            dname = tmp.mkdtemp()

            # Make sure we can run python files as if they we
            # executables (an issue on Windows)
            fname = os.path.join(dname, 'test.py')
            with open(fname, 'w') as F:
                F.write(f"#!{sys.executable}\nimport sysn\n")
            os.chmod(fname, 0o755)
            try:
                subprocess.run([fname])
            except OSError:
                raise unittest.SkipTest(
                    "python scripts are not registered as executable"
                )

            fname = os.path.join(dname, 'test_rc.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('HERE\\n')
sys.exit(2)
""")
            os.chmod(fname, 0o755)
            opt.config.executable = fname
            with LoggingIntercept() as LOG:
                self.assertIsNone(opt.version())
                self.assertEqual(opt.available(), Availability.NotFound)
            self.assertEqual(
                "Failed running GAMS command to get version (non-zero returncode 2):\n"
                "    FAIL\n    HERE\n",
                LOG.getvalue(),
            )

            fname = os.path.join(dname, 'test_bad_ver.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('GAMS Release : 26.1\\n')
sys.stdout.write('HERE\\n')
""")
            os.chmod(fname, 0o755)
            opt.config.executable = fname
            with LoggingIntercept() as LOG:
                self.assertIsNone(opt.version())
                self.assertEqual(opt.available(), Availability.NotFound)
            self.assertEqual(
                "Failed parsing GAMS version (version not found while parsing):\n"
                "    FAIL\n    GAMS Release : 26.1\n    HERE\n",
                LOG.getvalue(),
            )

            fname = os.path.join(dname, 'test_demo.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('*** GAMS Release     : 45.1.0 88bbff72\\n')
sys.stdout.write('*** GAMS Demo, for EULA and demo limitations\\n')
sys.stdout.write('HERE\\n')
""")
            os.chmod(fname, 0o755)
            opt.config.executable = fname
            with LoggingIntercept() as LOG:
                self.assertEqual(opt.version(), (45, 1, 0))
                self.assertEqual(opt.available(), Availability.LimitedLicense)
            self.assertEqual("", LOG.getvalue())

            fname = os.path.join(dname, 'test_full.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('*** GAMS Release     : 45.1.0 88bbff72\\n')
sys.stdout.write('*** Evaluation expiration date (GAMS base module)\\n')
sys.stdout.write('HERE\\n')
""")
            os.chmod(fname, 0o755)
            opt.config.executable = fname
            with LoggingIntercept() as LOG:
                self.assertEqual(opt.version(), (45, 1, 0))
                self.assertEqual(opt.available(), Availability.FullLicense)
            self.assertEqual("", LOG.getvalue())

    def test_write_gms_file(self):
        # We are creating a simple model with 1 variable to check for gams execution
        opt = gams.GAMS()
        config = gams.GAMSConfig()
        result = self._run_simple_model(config, 1)
        self.assertTrue(result)

        # Pass it some options that ARE on the command line and create a .gms file
        # Currently solver_options is not implemented in the new interface
        solver_exec = config.executable.path()
        opt = gams.GAMS(solver_options={'iterLim': 1})
        with TempfileManager.new_context() as temp:
            dname = temp.mkdtemp()
            if not os.path.exists(dname):
                os.mkdir(dname)
            filename = os.path.join(dname, 'test.gms')
            with open(filename, 'w') as FILE:
                FILE.write(self._simple_model(1))
            result = subprocess.run(
                [solver_exec, filename, "curdir=" + dname, 'lo=0'],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            self.assertTrue(result.returncode == 0)
            self.assertTrue(os.path.isfile(filename))


@unittest.skipIf(not gams_available, "The 'gams' command is not available")
@unittest.pytest.mark.solver("gams")
class TestGAMS(unittest.TestCase):
    def create_model(self):
        m = pyo.ConcreteModel('TestModel')
        m.x = pyo.Var(initialize=1.5, bounds=(-5, 5))
        m.y = pyo.Var(initialize=1.5, bounds=(-5, 5))

        m.obj = pyo.Objective(
            expr=(1.0 - m.x) + 100.0 * (m.y - m.x), sense=pyo.minimize
        )
        return m

    def create_infeasible_model(self):
        m = pyo.ConcreteModel('TestModel')
        m.x = pyo.Var(initialize=1.5, bounds=(-4, 4))
        m.y = pyo.Var(initialize=1.5, bounds=(-4, 4))

        m.c1 = pyo.Constraint(expr=m.x + m.y >= 5)
        m.c2 = pyo.Constraint(expr=-m.x + m.y >= 5)
        m.obj = pyo.Objective(rule=m.x + m.y, sense=pyo.minimize)
        return m

    def test_gams_config(self):
        # Test default initialization
        config = gams.GAMSConfig()
        self.assertTrue(config.load_solutions)
        self.assertIsInstance(config.solver_options, ConfigDict)
        self.assertIsInstance(config.executable, ExecutableData)

        # Test custom initialization
        solver = SolverFactory('gams_v2', executable='/path/to/exe')
        self.assertFalse(solver.config.tee)
        self.assertIsNone(solver.config.executable.path())
        self.assertTrue(solver.config.executable._registered_name.startswith('/path'))

    @unittest.skipIf(not gams.gdxcc_available, "'gdx' requires the gdx/gdxcc module")
    def test_gams_solve_gdx(self):
        # Gut check - does it solve?
        model = self.create_model()
        result = gams.GAMS().solve(model, writer_config={'put_results_format': 'gdx'})
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, -1004)
        self.assertAlmostEqual(model.x.value, 5)
        self.assertAlmostEqual(model.y.value, -5)

        model.obj.sense = pyo.maximize
        result = gams.GAMS().solve(model, writer_config={'put_results_format': 'gdx'})
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, 1006)
        self.assertAlmostEqual(model.x.value, -5)
        self.assertAlmostEqual(model.y.value, 5)

    def test_gams_solve_dat(self):
        # Gut check - does it solve?
        model = self.create_model()
        result = gams.GAMS().solve(model, writer_config={'put_results_format': 'dat'})
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, -1004)
        self.assertAlmostEqual(model.x.value, 5)
        self.assertAlmostEqual(model.y.value, -5)

        model.obj.sense = pyo.maximize
        result = gams.GAMS().solve(model, writer_config={'put_results_format': 'dat'})
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, 1006)
        self.assertAlmostEqual(model.x.value, -5)
        self.assertAlmostEqual(model.y.value, 5)

    @unittest.skipIf(not gams.gdxcc_available, "'gdx' requires the gdx/gdxcc module")
    def test_gams_solve_noload_gdx(self):
        # Gut check - does it solve?
        model = self.create_model()
        result = gams.GAMS().solve(
            model, load_solutions=False, writer_config={'put_results_format': 'gdx'}
        )
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, -1004)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

        model.obj.sense = pyo.maximize
        result = gams.GAMS().solve(
            model, load_solutions=False, writer_config={'put_results_format': 'gdx'}
        )
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, 1006)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

    def test_gams_solve_noload_dat(self):
        # Gut check - does it solve?
        model = self.create_model()
        result = gams.GAMS().solve(
            model, load_solutions=False, writer_config={'put_results_format': 'dat'}
        )
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, -1004)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

        model.obj.sense = pyo.maximize
        result = gams.GAMS().solve(
            model, load_solutions=False, writer_config={'put_results_format': 'dat'}
        )
        self.assertEqual(
            result.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(result.solution_status, SolutionStatus.optimal)
        self.assertAlmostEqual(result.incumbent_objective, 1006)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

    @unittest.skipIf(not gams.gdxcc_available, "'gdx' requires the gdx/gdxcc module")
    def test_gams_solve_infeasible_gdx(self):
        # Gut check - does it solve?
        model = self.create_infeasible_model()
        result = gams.GAMS().solve(
            model,
            raise_exception_on_nonoptimal_result=False,
            writer_config={'put_results_format': 'gdx'},
        )
        self.assertEqual(
            result.termination_condition, TerminationCondition.provenInfeasible
        )
        self.assertEqual(result.solution_status, SolutionStatus.infeasible)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

        model.obj.sense = pyo.maximize
        with self.assertRaisesRegex(
            NoOptimalSolutionError, "Solver did not find the optimal solution."
        ):
            result = gams.GAMS().solve(
                model, writer_config={'put_results_format': 'gdx'}
            )
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

    def test_gams_solve_infeasible_dat(self):
        # Gut check - does it solve?
        model = self.create_infeasible_model()
        result = gams.GAMS().solve(
            model,
            raise_exception_on_nonoptimal_result=False,
            writer_config={'put_results_format': 'dat'},
        )
        self.assertEqual(
            result.termination_condition, TerminationCondition.provenInfeasible
        )
        self.assertEqual(result.solution_status, SolutionStatus.infeasible)
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

        model.obj.sense = pyo.maximize
        with self.assertRaisesRegex(
            NoOptimalSolutionError, "Solver did not find the optimal solution."
        ):
            result = gams.GAMS().solve(
                model, writer_config={'put_results_format': 'dat'}
            )
        self.assertAlmostEqual(model.x.value, 1.5)
        self.assertAlmostEqual(model.y.value, 1.5)

    def test_bad_gams_executable(self):
        model = self.create_model()
        solver = gams.GAMS()
        solver.config.executable = "/path/to/bogus/gams"

        with self.assertRaisesRegex(
            ApplicationError, "GAMS: 'gams' executable not found"
        ):
            solver.solve(model)

        # Now try pointing to "executables" that are not GAMS
        with TempfileManager as tmp:
            dname = tmp.mkdtemp()
            solver.config.working_dir = dname

            # Make sure we can run python files as if they we
            # executables (an issue on Windows)
            fname = os.path.join(dname, 'test.py')
            with open(fname, 'w') as F:
                F.write(f"#!{sys.executable}\nimport sys\n")
            try:
                subprocess.run([fname])
            except OSError:
                raise unittest.SkipTest(
                    "python scripts are not registered as executable"
                )

            fname = os.path.join(dname, 'test_rc.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
lst = sys.argv[1].replace('.gms', '.lst')
for arg in sys.argv:
    if arg.startswith('o='):
        lst = arg[2:]
with open(lst, 'w') as LST:
    LST.write("Here is a multiline\\nlisting file.\\n")
sys.stderr.write('FAIL\\n')
sys.stdout.write('HERE\\n')
sys.exit(2)
""")
            os.chmod(fname, 0o755)
            solver.config.executable = fname
            with LoggingIntercept() as LOG:
                with self.assertRaisesRegex(
                    RuntimeError,
                    r"GAMS process encountered an error \(returncode=2\).\n"
                    r"Check listing file for details.",
                ):
                    solver.solve(model)
            self.assertEqual(
                "\n".join(
                    [
                        "GAMS process encountered an error (returncode=2).",
                        "Check listing file for details.",
                        "",
                        "FAIL",
                        "HERE",
                        "",
                        "GAMS Listing file:",
                        "",
                        "Here is a multiline",
                        "listing file.",
                    ]
                )
                + "\n",
                LOG.getvalue(),
            )

        with TempfileManager as tmp:
            dname = tmp.mkdtemp()
            solver.config.working_dir = dname

            fname = os.path.join(dname, 'test_rc_no_lst.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('HERE\\n')
sys.exit(2)
""")
            os.chmod(fname, 0o755)
            solver.config.executable = fname
            with LoggingIntercept() as LOG:
                with self.assertRaisesRegex(
                    RuntimeError,
                    r"GAMS process encountered an error \(returncode=2\).\n"
                    r"Check listing file for details.",
                ):
                    solver.solve(model)
            self.assertEqual(
                "\n".join(
                    [
                        "GAMS process encountered an error (returncode=2).",
                        "Check listing file for details.",
                        "",
                        "FAIL",
                        "HERE",
                    ]
                )
                + "\n",
                LOG.getvalue(),
            )

        with TempfileManager as tmp:
            dname = tmp.mkdtemp()
            solver.config.working_dir = dname

            fname = os.path.join(dname, 'test_rc3_no_lst.py')
            with open(fname, 'w') as F:
                F.write(f"""\
#!{sys.executable}
import sys
sys.stderr.write('FAIL\\n')
sys.stdout.write('HERE\\n')
sys.exit(3)
""")
            os.chmod(fname, 0o755)
            solver.config.executable = fname
            with LoggingIntercept() as LOG:
                with self.assertRaisesRegex(
                    RuntimeError,
                    r"GAMS process encountered an error \(returncode=3\).\n"
                    r"Error rc=3 \(GAMS execution error\), to be determined later.\n"
                    r"Check listing file for details.",
                ):
                    solver.solve(model)
            self.assertEqual(
                "\n".join(
                    [
                        "GAMS process encountered an error (returncode=3).",
                        "Error rc=3 (GAMS execution error), to be determined later.",
                        "Check listing file for details.",
                        "",
                        "FAIL",
                        "HERE",
                    ]
                )
                + "\n",
                LOG.getvalue(),
            )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import (
    TransformationFactory,
    Var,
    NonNegativeReals,
    Constraint,
    Objective,
    Block,
    value,
)

from pyomo.common.modeling import unique_component_name
from pyomo.core.plugins.transform.hierarchy import NonIsomorphicTransformation
from pyomo.common.config import ConfigBlock, ConfigValue
from pyomo.core.base import ComponentUID
from pyomo.common.deprecation import deprecation_warning


def target_list(x):
    deprecation_msg = (
        "In future releases ComponentUID targets will no "
        "longer be supported in the core.add_slack_variables "
        "transformation. Specify targets as a Constraint or "
        "list of Constraints."
    )
    if isinstance(x, ComponentUID):
        if deprecation_msg:
            deprecation_warning(deprecation_msg, version='5.7.1')
            # only emit the message once
            deprecation_msg = None
        # [ESJ 07/15/2020] We have to just pass it through because we need the
        # instance in order to be able to do anything about it...
        return [x]
    elif getattr(x, 'ctype', None) is Constraint:
        return [x]
    elif hasattr(x, '__iter__'):
        ans = []
        for i in x:
            if isinstance(i, ComponentUID):
                if deprecation_msg:
                    deprecation_warning(deprecation_msg, version='5.7.1')
                    deprecation_msg = None
                # same as above...
                ans.append(i)
            elif getattr(i, 'ctype', None) is Constraint:
                ans.append(i)
            else:
                raise ValueError(
                    "Expected Constraint or list of Constraints."
                    "\n\tReceived %s" % (type(i),)
                )
        return ans
    else:
        raise ValueError(
            "Expected Constraint or list of Constraints.\n\tReceived %s" % (type(x),)
        )


import logging

logger = logging.getLogger('pyomo.core')


@TransformationFactory.register(
    'core.add_slack_variables',
    doc="Create a model where we add slack variables to every constraint "
    "and add new objective penalizing the sum of the slacks",
)
class AddSlackVariables(NonIsomorphicTransformation):
    """
    This plugin adds slack variables to every constraint or to the constraints
    specified in targets.
    """

    CONFIG = ConfigBlock("core.add_slack_variables")
    CONFIG.declare(
        'targets',
        ConfigValue(
            default=None,
            domain=target_list,
            description="target or list of targets to which slacks will be added",
            doc="This specifies the list of Constraints to add slack variables to.",
        ),
    )

    def __init__(self, **kwds):
        kwds['name'] = "add_slack_vars"
        super(AddSlackVariables, self).__init__(**kwds)

    def _apply_to(self, instance, **kwds):
        self._apply_to_impl(instance, **kwds)

    def _apply_to_impl(self, instance, **kwds):
        config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)
        targets = config.targets

        if targets is None:
            constraintDatas = instance.component_data_objects(
                Constraint, descend_into=True
            )
        else:
            constraintDatas = []
            for t in targets:
                if isinstance(t, ComponentUID):
                    cons = t.find_component(instance)
                    if cons.is_indexed():
                        for i in cons:
                            constraintDatas.append(cons[i])
                    else:
                        constraintDatas.append(cons)
                else:
                    # we know it's a constraint because that's all we let
                    # through the config block validation.
                    if t.is_indexed():
                        for i in t:
                            constraintDatas.append(t[i])
                    else:
                        constraintDatas.append(t)

        # deactivate the objective
        for o in instance.component_data_objects(Objective):
            o.deactivate()

        # create block where we can add slack variables safely
        xblockname = unique_component_name(instance, "_core_add_slack_variables")
        instance.add_component(xblockname, Block())
        xblock = instance.component(xblockname)

        obj_expr = 0
        for cons in constraintDatas:
            if (cons.lower is not None and cons.upper is not None) and value(
                cons.lower
            ) > value(cons.upper):
                # this is a structural infeasibility so slacks aren't going to
                # help:
                raise RuntimeError(
                    "Lower bound exceeds upper bound in constraint %s" % cons.name
                )
            if not cons.active:
                continue
            cons_name = cons.getname(fully_qualified=True)
            lower = cons.lower
            body = cons.body
            upper = cons.upper
            if lower is not None:
                # we add positive slack variable to body:
                # declare positive slack
                varName = "_slack_plus_" + cons_name
                posSlack = Var(within=NonNegativeReals)
                xblock.add_component(varName, posSlack)
                # add positive slack to body expression
                body += posSlack
                # penalize slack in objective
                obj_expr += posSlack
            if upper is not None:
                # we subtract a positive slack variable from the body:
                # declare slack
                varName = "_slack_minus_" + cons_name
                negSlack = Var(within=NonNegativeReals)
                xblock.add_component(varName, negSlack)
                # add negative slack to body expression
                body -= negSlack
                # add slack to objective
                obj_expr += negSlack
            cons.set_value((lower, body, upper))
        # make a new objective that minimizes sum of slack variables
        xblock._slack_objective = Objective(expr=obj_expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

"""
This model is an adaptation of Eason & Biegler's example in the
original version of the Trust Region solver.

Eason, J.P. and Biegler, L.T. (2018), Advanced trust region optimization
strategies for glass box/black box models. AIChE J, 64: 3934-3943.
https://doi.org/10.1002/aic.16364
"""

from pyomo.environ import (
    ConcreteModel,
    Var,
    Reals,
    ExternalFunction,
    sin,
    cos,
    sqrt,
    Constraint,
    Objective,
)
from pyomo.opt import SolverFactory


def ext_fcn(a, b):
    return sin(a - b)


def grad_ext_fcn(args, fixed):
    a, b = args[:2]
    return [cos(a - b), -cos(a - b)]


def create_model():
    m = ConcreteModel()
    m.name = 'Example 1: Eason'
    m.z = Var(range(3), domain=Reals, initialize=2.0)
    m.x = Var(range(2), initialize=2.0)
    m.x[1] = 1.0

    m.ext_fcn = ExternalFunction(ext_fcn, grad_ext_fcn)

    m.obj = Objective(
        expr=(m.z[0] - 1.0) ** 2
        + (m.z[0] - m.z[1]) ** 2
        + (m.z[2] - 1.0) ** 2
        + (m.x[0] - 1.0) ** 4
        + (m.x[1] - 1.0) ** 6
    )

    m.c1 = Constraint(
        expr=m.x[0] * m.z[0] ** 2 + m.ext_fcn(m.x[0], m.x[1]) == 2 * sqrt(2.0)
    )
    m.c2 = Constraint(expr=m.z[2] ** 4 * m.z[1] ** 2 + m.z[1] == 8 + sqrt(2.0))
    return m


def main():
    m = create_model()
    optTRF = SolverFactory('trustregion', maximum_iterations=10, verbose=True)
    optTRF.solve(m, [m.z[0], m.z[1], m.z[2]])


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
from pyomo.environ import (
    ConcreteModel,
    Param,
    RangeSet,
    Var,
    Reals,
    Binary,
    PositiveIntegers,
)


def _cost_rule(model, n, m):
    # We will assume costs are an arbitrary function of the indices
    return math.sin(n * 2.33333 + m * 7.99999)


def create_model(n=3, m=3, p=2):
    model = ConcreteModel(name="M1")

    model.N = Param(initialize=n, within=PositiveIntegers)
    model.M = Param(initialize=m, within=PositiveIntegers)
    model.P = Param(initialize=p, within=RangeSet(1, model.N), mutable=True)

    model.Locations = RangeSet(1, model.N)
    model.Customers = RangeSet(1, model.M)

    model.cost = Param(
        model.Locations, model.Customers, initialize=_cost_rule, within=Reals
    )
    model.serve_customer_from_location = Var(
        model.Locations, model.Customers, bounds=(0.0, 1.0)
    )
    model.select_location = Var(model.Locations, within=Binary)

    @model.Objective()
    def obj(model):
        return sum(
            model.cost[n, m] * model.serve_customer_from_location[n, m]
            for n in model.Locations
            for m in model.Customers
        )

    @model.Constraint(model.Customers)
    def single_x(model, m):
        return (
            sum(model.serve_customer_from_location[n, m] for n in model.Locations)
            == 1.0
        )

    @model.Constraint(model.Locations, model.Customers)
    def bound_y(model, n, m):
        return model.serve_customer_from_location[n, m] <= model.select_location[n]

    @model.Constraint()
    def num_facilities(model):
        return sum(model.select_location[n] for n in model.Locations) == model.P

    return model



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Tests the Bounds Tightening module."""

import pyomo.common.unittest as unittest
from pyomo.environ import ConcreteModel, Constraint, TransformationFactory, Var, value


class TestIntervalTightener(unittest.TestCase):
    """Tests Bounds Tightening."""

    def test_constraint_bound_tightening(self):
        # Check for no coefficients
        m = ConcreteModel()
        m.v1 = Var(initialize=7, bounds=(7, 10))
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=m.v1 >= m.v2 + m.v3 + m.v4 + 1)

        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), 0)
        self.assertEqual(value(m.c1.lower), 0)

    def test_less_than_constraint(self):
        m = ConcreteModel()
        m.v1 = Var(initialize=7, bounds=(7, 10))
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=m.v1 <= m.v2 + m.v3 + m.v4)

        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), 0)
        self.assertEqual(value(m.c1.lower), -8)

    def test_constraint_with_coef(self):
        """Test with coefficient on constraint."""
        m = ConcreteModel()
        m.v1 = Var(initialize=7, bounds=(7, 10))
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=m.v1 <= 2 * m.v2 + m.v3 + m.v4)

        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), -1)
        self.assertEqual(value(m.c1.lower), -13)

    def test_unbounded_var(self):
        """test with unbounded variables"""
        m = ConcreteModel()
        m.v1 = Var(initialize=7)
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=m.v1 <= 2 * m.v2 + m.v3 + m.v4)

        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())

    def test_unbounded_one_direction(self):
        """Unbounded in one direction"""
        m = ConcreteModel()
        m.v1 = Var(initialize=7, bounds=(-float('inf'), 10))
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=m.v1 <= 2 * m.v2 + m.v3 + m.v4)

        self.assertEqual(value(m.c1.upper), 0)
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), -1)
        self.assertFalse(m.c1.has_lb())

    def test_negative_coeff(self):
        """Unbounded in one direction with negative coefficient"""
        m = ConcreteModel()
        m.v1 = Var(initialize=7, bounds=(1, float('inf')))
        m.v2 = Var(initialize=2, bounds=(2, 5))
        m.v3 = Var(initialize=6, bounds=(6, 9))
        m.v4 = Var(initialize=1, bounds=(1, 1))
        m.c1 = Constraint(expr=2 * m.v2 + m.v3 + m.v4 - m.v1 <= 50)

        self.assertEqual(value(m.c1.upper), 50)
        self.assertTrue(m.c1.has_ub())
        self.assertFalse(m.c1.has_lb())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.upper), 19)
        self.assertFalse(m.c1.has_lb())

    def test_ignore_nonlinear(self):
        m = ConcreteModel()
        m.v1 = Var()
        m.c1 = Constraint(expr=m.v1 * m.v1 >= 2)

        self.assertEqual(value(m.c1.lower), 2)
        self.assertFalse(m.c1.has_ub())
        TransformationFactory('core.tighten_constraints_from_vars').apply_to(m)
        self.assertEqual(value(m.c1.lower), 2)
        self.assertFalse(m.c1.has_ub())


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Main driver module for GDPopt solver.

22.5.13 changes:
- rewrite of all algorithms
- deprecate 'strategy' in favor of 'algorithm'
- deprecate 'init_strategy' in favor of 'init_algorithm'
20.2.28 changes:
- bugfixes on tests
20.1.22 changes:
- improved subsolver time limit support for GAMS interface
- add maxTimeLimit exit condition for GDPopt-LBB
- add token Big M for reactivated constraints in GDPopt-LBB
- activate fbbt for branch-and-bound nodes
20.1.15 changes:
- internal cleanup of codebase
- merge GDPbb capabilities (logic-based branch and bound)
- refactoring of GDPbb code
- update logging information to include subsolver options
- improve SuppressInfeasibleWarning
- simplify mip preprocessing
- remove not-fully-implemented 'backtracking' from LOA
19.10.11 changes:
- bugfix on SolverStatus error message
19.5.13 changes:
- add handling to integer cuts for disjunct pruning during FBBT
19.4.23 changes:
- add support for linear subproblems
- use automatic differentiation for large constraints
- bugfixes on time limit support
- treat fixed variables as constants in GLOA cut generation
19.3.25 changes:
- add rudimentary time limit support
- start keeping basic changelog

"""

from pyomo.common.config import document_kwargs_from_configdict, ConfigDict
from pyomo.contrib.gdpopt import __version__
from pyomo.contrib.gdpopt.config_options import (
    _add_common_configs,
    _supported_algorithms,
    _get_algorithm_config,
)
from pyomo.opt.base import SolverFactory


def _handle_strategy_deprecation(config):
    # This method won't be needed when the strategy arg is removed, but for now,
    # we need to copy it over as algorithm. The config system already gave the
    # deprecation warning.
    if config.algorithm is None and config.strategy is not None:
        config.algorithm = config.strategy


@SolverFactory.register(
    'gdpopt',
    doc='The GDPopt decomposition-based '
    'Generalized Disjunctive Programming (GDP) solver',
)
class GDPoptSolver:
    """Decomposition solver for Generalized Disjunctive Programming (GDP)
    problems.

    The GDPopt (Generalized Disjunctive Programming optimizer) solver applies a
    variety of decomposition-based approaches to solve Generalized Disjunctive
    Programming (GDP) problems. GDP models can include nonlinear, continuous
    variables and constraints, as well as logical conditions.

    These approaches include:

    - Logic-based outer approximation (LOA)
    - Logic-based branch-and-bound (LBB)
    - Partial surrogate cuts [pending]
    - Generalized Bender decomposition [pending]

    This solver implementation was developed by Carnegie Mellon University in
    the research group of Ignacio Grossmann.

    For nonconvex problems, LOA may not report rigorous lower/upper bounds.

    Questions: Please make a post at StackOverflow and/or contact Qi Chen
    <https://github.com/qtothec> or David Bernal <https://github.com/bernalde>.

    Several key GDPopt components were prototyped by BS and MS students:

    - Logic-based branch and bound: Sunjeev Kale
    - MC++ interface: Johnny Bates
    - LOA set-covering initialization: Eloy Fernandez
    - Logic-to-linear transformation: Romeo Valentin

    """

    CONFIG = ConfigDict("GDPopt")
    _add_common_configs(CONFIG)

    @document_kwargs_from_configdict(CONFIG)
    def solve(self, model, **kwds):
        """Solve the model.

        Args:
            model (Block): a Pyomo model or block to be solved

        """
        # The algorithm should have been specified as an argument to the solve
        # method. We will instantiate an ephemeral instance of the correct
        # solver and call its solve method.
        options = kwds.pop('options', {})
        config = self.CONFIG(options, preserve_implicit=True)
        # Don't complain about extra things, they aren't for us. We just need to
        # get the algorithm and then our job is done.
        config.set_value(kwds, skip_implicit=True)

        alg_config = _get_algorithm_config()(options, preserve_implicit=True)
        alg_config.set_value(kwds, skip_implicit=True)

        _handle_strategy_deprecation(alg_config)
        algorithm = alg_config.algorithm
        if algorithm is None:
            raise ValueError(
                "No algorithm was specified to the solve method. "
                "Please specify an algorithm or use an "
                "algorithm-specific solver."
            )

        # get rid of 'algorithm' and 'strategy' if they exist so that the solver
        # can validate.
        kwds.pop('algorithm', None)
        kwds.pop('strategy', None)

        # The algorithm has already been validated, so this will work.
        return SolverFactory(_supported_algorithms[algorithm][0]).solve(model, **kwds)

    # Support use as a context manager under current solver API
    def __enter__(self):
        return self

    def __exit__(self, t, v, traceback):
        pass

    def available(self, exception_flag=True):
        """Solver is always available. Though subsolvers may not be, they will
        raise an error when the time comes.
        """
        return True

    def license_is_valid(self):
        return True

    def version(self):
        """Return a 3-tuple describing the solver version."""
        return __version__

    _metasolver = False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


class _StaleFlagManager:
    def __init__(self):
        self._current = 0
        self.mark_all_as_stale()

    def stale_mapper(self, encode, value):
        if encode:
            return self.is_stale(value)
        else:
            if value:
                return 0
            else:
                return self.get_flag(0)

    def _get_flag(self, current_flag):
        """Return the current global stale flag value"""
        return self._current

    def _get_flag_delayed(self, current_flag):
        """Implement the "delayed" advancement of the global stale flag value

        This will continue to return the current value of the state flag
        until the first non-stale variable is updated (that it, it is
        passed the current stale flag when called).  This allows for
        updating stale variable values without incrementing the global
        stale flag, but will mark everything as stale as soon as a
        non-stale variable value is changed.

        """
        if current_flag == self._current:
            self._current += 1
            setattr(self, 'get_flag', getattr(self, '_get_flag'))
        return self._current

    def is_stale(self, val):
        """Return ``True`` if the passed value indicated a stale variable"""
        return val != self._current

    def mark_all_as_stale(self, delayed=False):
        """Advance the global stale flag, marking all variables as stale

        This is generally called immediately before and after a batch
        variable update (i.e. loading values from a solver result or
        stored solution).  Before the batch update
        :meth:`mark_all_as_stale` is called with ``delayed=False``,
        which immediately marks all variables as stale.  After the batch
        update, :meth:`mark_all_as_stale` is typically called with
        ``delayed=True``.  This allows additional stale variables to be
        updated without advancing the global flag, but as soon as any
        non-stale variable has its value changed, then the flag is
        advanced and all other variables become stale.

        """
        if delayed:
            setattr(self, 'get_flag', getattr(self, '_get_flag_delayed'))
        else:
            setattr(self, 'get_flag', getattr(self, '_get_flag'))
            self._current += 1


StaleFlagManager = _StaleFlagManager()
